"""
websocket - WebSocket client library for Python

Copyright (C) 2010 Hiroki Ohtani(liris)

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor,
    Boston, MA 02110-1335  USA

"""
import six

__all__ = ['NoLock', 'validate_utf8', 'extract_err_message', 'extract_error_code']


class NoLock(object):
    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_value, traceback):
        pass


try:
    # If wsaccel is available we use compiled routines to validate UTF-8
    # strings.
    from wsaccel.utf8validator import Utf8Validator

    def _validate_utf8(utfbytes):
        return Utf8Validator().validate(utfbytes)[0]

except ImportError:
    # UTF-8 validator
    # python implementation of http://bjoern.hoehrmann.de/utf-8/decoder/dfa/

    _UTF8_ACCEPT = 0
    _UTF8_REJECT = 12

    _UTF8D = [
        # The first part of the table maps bytes to character classes that
        # to reduce the size of the transition table and create bitmasks.
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        9,
        9,
        9,
        9,
        9,
        9,
        9,
        9,
        9,
        9,
        9,
        9,
        9,
        9,
        9,
        9,
        7,
        7,
        7,
        7,
        7,
        7,
        7,
        7,
        7,
        7,
        7,
        7,
        7,
        7,
        7,
        7,
        7,
        7,
        7,
        7,
        7,
        7,
        7,
        7,
        7,
        7,
        7,
        7,
        7,
        7,
        7,
        7,
        8,
        8,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        2,
        10,
        3,
        3,
        3,
        3,
        3,
        3,
        3,
        3,
        3,
        3,
        3,
        3,
        4,
        3,
        3,
        11,
        6,
        6,
        6,
        5,
        8,
        8,
        8,
        8,
        8,
        8,
        8,
        8,
        8,
        8,
        8,

        # The second part is a transition table that maps a combination
        # of a state of the automaton and a character class to a state.
        0,
        12,
        24,
        36,
        60,
        96,
        84,
        12,
        12,
        12,
        48,
        72,
        12,
        12,
        12,
        12,
        12,
        12,
        12,
        12,
        12,
        12,
        12,
        12,
        12,
        0,
        12,
        12,
        12,
        12,
        12,
        0,
        12,
        0,
        12,
        12,
        12,
        24,
        12,
        12,
        12,
        12,
        12,
        24,
        12,
        24,
        12,
        12,
        12,
        12,
        12,
        12,
        12,
        12,
        12,
        24,
        12,
        12,
        12,
        12,
        12,
        24,
        12,
        12,
        12,
        12,
        12,
        12,
        12,
        24,
        12,
        12,
        12,
        12,
        12,
        12,
        12,
        12,
        12,
        36,
        12,
        36,
        12,
        12,
        12,
        36,
        12,
        12,
        12,
        12,
        12,
        36,
        12,
        36,
        12,
        12,
        12,
        36,
        12,
        12,
        12,
        12,
        12,
        12,
        12,
        12,
        12,
        12,
    ]

    def _decode(state, codep, ch):
        tp = _UTF8D[ch]

        codep = (ch & 0x3f) | (codep << 6) if (state != _UTF8_ACCEPT) else (0xff >> tp) & ch
        state = _UTF8D[256 + state + tp]

        return state, codep

    def _validate_utf8(utfbytes):
        state = _UTF8_ACCEPT
        codep = 0
        for i in utfbytes:
            if six.PY2:
                i = ord(i)
            state, codep = _decode(state, codep, i)
            if state == _UTF8_REJECT:
                return False

        return True


def validate_utf8(utfbytes):
    """
    validate utf8 byte string.
    utfbytes: utf byte string to check.
    return value: if valid utf8 string, return true. Otherwise, return false.
    """
    return _validate_utf8(utfbytes)


def extract_err_message(exception):
    if exception.args:
        return exception.args[0]
    else:
        return None


def extract_error_code(exception):
    if exception.args and len(exception.args) > 1:
        return exception.args[0] if isinstance(exception.args[0], int) else None



#!/usr/bin/env python
# dtuple.py
#
# Copyright (C) 2007-2008 Francois Aucamp, Meraka Institute, CSIR
# See AUTHORS for all authors and contact information.
#
# License: GNU Lesser General Public License, version 3 or later; see COPYING
#          included in this archive for details.
#
# This library is free software, distributed under the terms of
# the GNU Lesser General Public License Version 3, or any later version.
# See the COPYING file included in this archive
#
# The docstrings in this module contain epytext markup; API documentation
# may be created by processing this file with epydoc: http://epydoc.sf.net

from __future__ import absolute_import
from __future__ import print_function
from io import open
import six.moves.cPickle

import hashlib

from twisted.internet import defer

from .kademlia.node import rpcmethod  # @UnresolvedImport
from .node import EntangledNode  # @UnresolvedImport


class DistributedTupleSpacePeer(EntangledNode):
    """
    A specialized form of an Entangled DHT node that provides an API for
    participating in a distributed Tuple Space (aka Object Space)
    """
    def __init__(self, udpPort=4000, dataStore=None, routingTable=None, networkProtocol=None, **kwargs):
        EntangledNode.__init__(self, udpPort, dataStore, routingTable, networkProtocol, **kwargs)
        self._blockingGetRequests = {}
        self._blockingReadRequests = {}

    def put(self, dTuple, originalPublisherID=None):
        """
        Produces a tuple, and writes it into tuple space.

        @note: This method is generally called "out" in tuple space literature,
               but is renamed to "put" in this implementation to match the
               renamed "in"/"get" method (see the description for C{get()}).

        @param dTuple: The tuple to write into the distributed tuple space (it
                       is named "dTuple" to avoid a conflict with the Python
                       C{tuple} data type).
        @type dTuple: tuple

        @rtype: twisted.internet.defer.Deferred
        """

        # Look for any active listener tuples for the tuple we're about to publish
        listenerNodeID = []
        listenerKey = []

        def publishToTupleSpace(result):
            if result != 'get':
                # Extract "keywords" from the tuple
                subtupleKeys = self._keywordHashesFromTuple(dTuple)
                # Write the tuple to the DHT Tuple Space...
                h = hashlib.sha1()
                tupleValue = six.moves.cPickle.dumps(dTuple)
                h.update('tuple:' + tupleValue)
                mainKey = h.digest()
                self.iterativeStore(mainKey, tupleValue, originalPublisherID=originalPublisherID)
                # ...and now make it searchable, by writing the subtuples
                df = self._addToInvertedIndexes(subtupleKeys, mainKey)
                return df

        def sendTupleToNode(nodes):
            if listenerNodeID[0] in nodes:
                contact = nodes[nodes.index(listenerNodeID[0])]
                df = contact.receiveTuple(listenerKey[0], six.moves.cPickle.dumps(dTuple))
                return df

        def checkIfListenerExists(result):
            if result is not None:
                # The result will have a node ID and main listener tuple's key concatenated
                listenerNodeID.append(result[:20])  # 160 bits
                listenerKey.append(result[20:])
                # Another node is waiting for this tuple; we will send it the tuple directly
                listenerNodeID.append(listenerNodeID[0])
                # First remove the listener from the Tuple Space
                self.iterativeDelete(listenerKey[0])
                subtupleKeys = self._keywordHashesFromTuple(dTuple, True)
                self._removeFromInvertedIndexes(subtupleKeys, result)
                # ...now retrieve the contact for the target Node ID, and send it the tuple
                # TODO: perhaps ping this node to make sure its still active
                try:
                    contact = self._routingTable.getContact(listenerNodeID[0])
                except ValueError:
                    df = self.iterativeFindNode(listenerNodeID[0])
                    df.addCallback(sendTupleToNode)
                    df.addCallback(publishToTupleSpace)
                else:
                    # TODO: add a callback to this to determine if it was a read/get
                    df = contact.receiveTuple(listenerKey[0], six.moves.cPickle.dumps(dTuple))
                    df.addCallback(publishToTupleSpace)
            else:
                # Extract "keywords" from the tuple
                subtupleKeys = self._keywordHashesFromTuple(dTuple)
                # Write the tuple to the DHT Tuple Space...
                h = hashlib.sha1()
                tupleValue = six.moves.cPickle.dumps(dTuple)
                h.update('tuple:' + tupleValue)
                mainKey = h.digest()

                def putToSearchIndexes(result):
                    df = self._addToInvertedIndexes(subtupleKeys, mainKey)
                    return df

                df = self.iterativeStore(mainKey, tupleValue, originalPublisherID=originalPublisherID)
                df.addCallback(putToSearchIndexes)
                # ...and now make it searchable, by writing the subtuples
                #df = self._addToInvertedIndexes(subtupleKeys, mainKey)
            return df

        df = self._findKeyForTemplate(dTuple, True)
        df.addCallback(checkIfListenerExists)
        return df

    def get(self, template):
        """
        Reads and removes (consumes) a tuple from the tuple space.

        @type template: tuple

        @note: This method is generally called "in" in tuple space literature,
               but is renamed to "get" in this implementation to avoid
               a conflict with the Python C{in} keyword.
        """
        outerDf = defer.Deferred()

        def addListener(result):
            if result is None:
                # The tuple does not exist (yet) - add a listener for it
                h = hashlib.sha1()
                listenerKey = 'listener:' + six.moves.cPickle.dumps(template)
                h.update(listenerKey)
                listenerKey = h.digest()
                # Extract "listener keywords" from the template
                subtupleKeys = self._keywordHashesFromTemplate(template, True)
                # ...now write the listener tuple(s) to the DHT Tuple Space
                if subtupleKeys is None:
                    # Deterministic template; all values are fully specified
                    self.iterativeStore(listenerKey, self.id + listenerKey)
                else:
                    self._addToInvertedIndexes(subtupleKeys, self.id + listenerKey)
                self._blockingGetRequests[listenerKey] = outerDf
            else:
                outerDf.callback(result)

        df = self.getIfExists(template)
        df.addCallback(addListener)
        return outerDf


#    def get(self, template):
#        """ Reads and removes (consumes) a tuple from the tuple space.
#
#        @type template: tuple
#
#        @note: This method is generally called "in" in tuple space literature,
#               but is renamed to "get" in this implementation to avoid
#               a conflict with the Python C{in} keyword.
#        """
#        outerDf = defer.Deferred()
#
#        mainTupleKey = []
#        def retrieveTupleValue(tupleKey):
#            if tupleKey == None:
#                # No tuple was found
#                return None
#            else:
#                mainTupleKey.append(tupleKey)
#                #TODO: kademlia is going to replicate the un-updated inverted index; stop that from happening!!
#                _df = self.iterativeFindValue(tupleKey)
#                _df.addCallback(returnTuple)
#                return _df
#
#        def returnTuple(value):
#            if type(value) == dict:
#                # tuple was found
#                tupleValue = value[mainTupleKey[0]]
#                # Remove the tuple itself from the DHT
#                self.iterativeDelete(mainTupleKey[0])
#                # Un-serialize the tuple...
#                dTuple = cPickle.loads(tupleValue)
#                # ...now remove all inverted index entries of the tuple, and return
#                subtupleKeys = self._keywordHashesFromTuple(dTuple)
#                self._removeFromInvertedIndexes(subtupleKeys, mainTupleKey[0])
#                return dTuple
#            else:
#                # tuple was not found
#                return None
#
#
#        def addListener(result):
#            if result == None:
#                # The tuple does not exist (yet) - add a listener for it
#                h = hashlib.sha1()
#                listenerKey = 'listener:' + cPickle.dumps(template)
#                h.update(listenerKey)
#                listenerKey = h.digest()
#                # Extract "listener keywords" from the template
#                subtupleKeys = self._keywordHashesFromTemplate(template, True)
#                # ...now write the listener tuple(s) to the DHT Tuple Space
#                if subtupleKeys == None:
#                    # Deterministic template; all values are fully specified
#                    self.iterativeStore(listenerKey, self.id + listenerKey)
#                else:
#                    self._addToInvertedIndexes(subtupleKeys, self.id + listenerKey)
#                self._blockingGetRequests[listenerKey] = outerDf
#            else:
#                outerDf.callback(result)
#
#        df = self._findKeyForTemplate(template)
#        df.addCallback(retrieveTupleValue)
#        df.addCallback(addListener)
#        return outerDf

    def getIfExists(self, template, getListenerTuple=False):
        """
        Reads and removes (consumes) a tuple from the tuple space.

        @type template: tuple

        @param getListenerTuple: If set to True, look for a I{listener tuple}
                                 for this template; this is typically used
                                 to remove event handlers.
        @type getListenerTuple: bool

        @note: This method is generally called "in" in tuple space literature,
               but is renamed to "get" in this implementation to avoid
               a conflict with the Python C{in} keyword.
        """
        outerDf = defer.Deferred()

        mainTupleKey = []

        def retrieveTupleValue(tupleKey):
            if tupleKey is None:
                # No tuple was found
                outerDf.callback(None)
            else:
                mainTupleKey.append(tupleKey)
                # We use the find algorithm directly so that kademlia does not replicate the key
                if tupleKey in self._dataStore:
                    _df = defer.Deferred()
                    _df.callback({tupleKey: self._dataStore[tupleKey]})
                else:
                    _df = self._iterativeFind(tupleKey, rpc='findValue')
                _df.addCallback(returnTuple)

        def returnTuple(value):
            if isinstance(value, dict):
                # tuple was found
                tupleValue = value[mainTupleKey[0]]
                # Remove the tuple itself from the DHT
                self.iterativeDelete(mainTupleKey[0])
                # Un-serialize the tuple...
                dTuple = six.moves.cPickle.loads(tupleValue)
                # ...now remove all inverted index entries of the tuple, and return
                subtupleKeys = self._keywordHashesFromTuple(dTuple)
                self._removeFromInvertedIndexes(subtupleKeys, mainTupleKey[0])
                outerDf.callback(dTuple)
            else:
                # tuple was not found
                outerDf.callback(None)

        df = self._findKeyForTemplate(template, getListenerTuple)
        df.addCallback(retrieveTupleValue)

        return outerDf

    def read(self, template):
        """
        Non-destructively reads a tuple in the tuple space.

        This operation is similar to "get" (or "in") in that the peer builds a
        template and waits for a matching tuple in the tuple space. Upon
        finding a matching tuple, however, it copies it, leaving the original
        tuple in the tuple space.

        @note: This method is named "rd" in some other implementations.
        """
        outerDf = defer.Deferred()

        def addListener(result):
            if result is None:
                # The tuple does not exist (yet) - add a listener for it
                h = hashlib.sha1()
                listenerKey = 'listener:' + six.moves.cPickle.dumps(template)
                h.update(listenerKey)
                listenerKey = h.digest()
                # Extract "listener keywords" from the template
                subtupleKeys = self._keywordHashesFromTemplate(template, True)
                # ...now write the listener tuple(s) to the DHT Tuple Space
                if subtupleKeys is None:
                    # Deterministic template; all values are fully specified
                    self.iterativeStore(listenerKey, self.id + listenerKey)
                else:
                    self._addToInvertedIndexes(subtupleKeys, self.id + listenerKey)
                self._blockingReadRequests[listenerKey] = outerDf
            else:
                outerDf.callback(result)

        df = self.readIfExists(template)
        df.addCallback(addListener)
        return outerDf

    def readIfExists(self, template):
        """
        Non-destructively reads a tuple in the tuple space.

        This operation is similar to "get" (or "in") in that the peer builds a
        template and waits for a matching tuple in the tuple space. Upon
        finding a matching tuple, however, it copies it, leaving the original
        tuple in the tuple space.

        @note: This method is named "rd" in some other implementations.
        """
        outerDf = defer.Deferred()
        mainTupleKey = []

        def retrieveTupleValue(tupleKey):
            if tupleKey is None:
                # No tuple was found
                outerDf.callback(None)
            else:
                mainTupleKey.append(tupleKey)
                _df = self.iterativeFindValue(tupleKey)
                _df.addCallback(returnTuple)

        def returnTuple(value):
            if isinstance(value, dict):
                # tuple was found
                tupleValue = value[mainTupleKey[0]]
                # Un-serialize the tuple
                dTuple = six.moves.cPickle.loads(tupleValue)
                outerDf.callback(dTuple)
            else:
                # tuple was not found
                outerDf.callback(None)

        df = self._findKeyForTemplate(template)
        df.addCallback(retrieveTupleValue)
        return outerDf

    def _findKeyForTemplate(self, template, listener=False):
        """
        Main search algorithm for C{get()} and C{read()}
        """
        if listener:
            prependStr = 'listener:'
        else:
            prependStr = 'tuple:'
        # Prepare a deferred result for this operation
        outerDf = defer.Deferred()
        if listener:
            subtupleKeys = self._keywordHashesFromTuple(template, listener)
        else:
            subtupleKeys = self._keywordHashesFromTemplate(template, listener)

        kwIndex = [-1]  # using a list for this counter because Python doesn't allow binding a new value to a name in an enclosing (non-global) scope
        havePossibleMatches = [False]
        filteredResults = []

        listenerResults = []
        listenerSubtupleSetCounter = [0]

        # TODO: If all elements in the template are None, we only have the tuple length... maybe raise an exception?

        def filterResult(result):
            kwKey = subtupleKeys[kwIndex[0]]
            if isinstance(result, dict):
                # Value was found; this should be list of keys for tuples matching this criterion
                index = result[kwKey]
                if havePossibleMatches[0] == False:
                    havePossibleMatches[0] = True
                    filteredResults.extend(index)
                else:
                    # Filter the our list of possible matching tuples with the new results
                    delKeys = []
                    for tupleKey in filteredResults:
                        if tupleKey not in index:
                            delKeys.append(tupleKey)
                    for tupleKey in delKeys:
                        filteredResults.remove(tupleKey)
                if len(filteredResults) == 0:
                    # No matches for this template exist at this point; there is no use in searching further
                    outerDf.callback(None)
                else:
                    findNextSubtuple()
            else:
                # Value wasn't found; thus no matches for this template exist - stop the search
                outerDf.callback(None)

        def filterListenerResult(result):
            """
            Same as filterResult(), except that 2 sets of subtuples keys'
            results are OR'ed.
            """
            if kwIndex[0] == -1:
                # This was the deterministic search
                if isinstance(result, dict):
                    # An exact template match was found, callback with this
                    outerDf.callback(result[mainKey])
                else:
                    # The deterministic search did not find anything; start searching subtuples
                    findNextSubtuple()
                return

            kwKey = subtupleKeys[kwIndex[0]]

            if isinstance(result, dict):
                # Value was found; this should be list of keys for tuples matching this criterion
                index = result[kwKey]
                listenerResults.extend(index)
            listenerSubtupleSetCounter[0] += 1

            if listenerSubtupleSetCounter[0] == 3:
                if havePossibleMatches[0] == False:
                    havePossibleMatches[0] = True
                    filteredResults.extend(listenerResults)
                else:
                    # Filter the our list of possible matching tuples with the new results
                    delKeys = []
                    for tupleKey in filteredResults:
                        if tupleKey not in listenerResults:
                            delKeys.append(tupleKey)
                    for tupleKey in delKeys:
                        try:
                            filteredResults.remove(tupleKey)
                        except ValueError:
                            pass

                if len(filteredResults) == 0:
                    # No matches for this template exist at this point; there is no use in searching further
                    outerDf.callback(None)
                else:
                    # Reset the cycle
                    listenerSubtupleSetCounter[0] = 0
                    while len(listenerResults):
                        listenerResults.pop()
                    findNextSubtuple()
            else:
                findNextSubtuple()

        def findNextSubtuple(results=None):
            kwIndex[0] += 1
            if kwIndex[0] < len(subtupleKeys):
                kwKey = subtupleKeys[kwIndex[0]]
                # TODO: kademlia is going to replicate the un-updated inverted index; stop that from happening!!
                df = self.iterativeFindValue(kwKey)
                if listener:
                    df.addCallback(filterListenerResult)
                else:
                    df.addCallback(filterResult)
            else:
                # We're done. Let the caller of the parent method know, and return the key of the first qualifying tuple in the list of results
                outerDf.callback(filteredResults[0])

        if subtupleKeys is None:
            # The template is deterministic; thus we can retrieve the corresponding tuple directly
            h = hashlib.sha1()
            tupleValue = six.moves.cPickle.dumps(template)
            h.update(prependStr + tupleValue)
            mainKey = h.digest()
            outerDf.callback(mainKey)
        else:
            if listener:
                # First look for an exact match if we are looking for listener tuples
                h = hashlib.sha1()
                tupleValue = six.moves.cPickle.dumps(template)
                h.update(prependStr + tupleValue)
                mainKey = h.digest()
                df = self.iterativeFindValue(mainKey)
                df.addCallback(filterListenerResult)
            else:
                # Query the DHT for the first subtuple (this implicitly specifies the requested tuple length as well)
                findNextSubtuple()

        return outerDf

    def _keywordHashesFromTuple(self, dTuple, listener=False):
        if listener:
            prependStr = 'listener:'
        else:
            prependStr = 'tuple:'
        subtupleKeys = []
        i = 0
        tupleLength = len(dTuple)
        for element in dTuple:
            # Because a tuple is immutable, the position of an element in the tuple
            # is a constant, along with the length of the tuple.
            # This causes each element in the tuble to have two identifiying subtuples (or "keywords")
            # which can be published to the DHT: tuple_length+position+data_type and tuple_length+position+data_value
            # (data_type is implicitly given by data_value)
            # TODO: with the current scheme, it is possible (but unlikely) that an ACTUAL tuple may clash with one of these subtuples
            typeSubtuple = (tupleLength, i, type(element))
            h = hashlib.sha1()
            h.update(prependStr + six.moves.cPickle.dumps(typeSubtuple))
            subtupleKeys.append(h.digest())
            valueSubtuple = (tupleLength, i, element)
            h = hashlib.sha1()
            h.update(prependStr + six.moves.cPickle.dumps(valueSubtuple))
            subtupleKeys.append(h.digest())
            wildcardSubtuple = (tupleLength, i, None)
            h = hashlib.sha1()
            h.update(prependStr + six.moves.cPickle.dumps(wildcardSubtuple))
            subtupleKeys.append(h.digest())
            i += 1
        return subtupleKeys

    def _keywordHashesFromTemplate(self, template, listener=False):
        if listener:
            prependStr = 'listener:'
        else:
            prependStr = 'tuple:'
        subtupleKeys = []
        i = 0
        tupleLength = len(template)
        deterministicElementCount = 0
        for element in template:
            # See the description in _keywordHashesFromTuple() for how these "keyword" subtuples are constructed
            # if element != None:
            # This element in the template describes the element's value or type
            if not isinstance(element, type) and element is not None:
                deterministicElementCount += 1
            subtuple = (tupleLength, i, element)
            h = hashlib.sha1()
            h.update(prependStr + six.moves.cPickle.dumps(subtuple))
            subtupleKeys.append(h.digest())
            # else:
            # The element is None; treat it as a wildcard
            #    pass
            i += 1
        if deterministicElementCount == tupleLength:
            # All of the elements are fully specified in this template; thus we can retrieve the corresponding tuple directly
            return None
        else:
            return subtupleKeys

    @rpcmethod
    def receiveTuple(self, listenerKey, pickledTuple):
        if listenerKey in self._blockingGetRequests:
            dTuple = six.moves.cPickle.loads(pickledTuple)
            df = self._blockingGetRequests[listenerKey]
            df.callback(dTuple)
            return 'get'
        elif listenerKey in self._blockingReadRequests:
            dTuple = six.moves.cPickle.loads(pickledTuple)
            df = self._blockingReadRequests[listenerKey]
            df.callback(dTuple)
            return 'read'

if __name__ == '__main__':
    import sys
    if len(sys.argv) < 2:
        print('Usage:\n%s UDP_PORT  [KNOWN_NODE_IP  KNOWN_NODE_PORT]' % sys.argv[0])
        print('or:\n%s UDP_PORT  [FILE_WITH_KNOWN_NODES]' % sys.argv[0])
        print('\nIf a file is specified, it should containg one IP address and UDP port\nper line, seperated by a space.')
        sys.exit(1)
    try:
        int(sys.argv[1])
    except ValueError:
        print('\nUDP_PORT must be an integer value.\n')
        print('Usage:\n%s UDP_PORT  [KNOWN_NODE_IP  KNOWN_NODE_PORT]' % sys.argv[0])
        print('or:\n%s UDP_PORT  [FILE_WITH_KNOWN_NODES]' % sys.argv[0])
        print('\nIf a file is specified, it should contain one IP address and UDP port\nper line, seperated by a space.')
        sys.exit(1)

    if len(sys.argv) == 4:
        knownNodes = [(sys.argv[2], int(sys.argv[3]))]
    elif len(sys.argv) == 3:
        knownNodes = []
        f = open(sys.argv[2], 'r')
        lines = f.readlines()
        f.close()
        for line in lines:
            ipAddress, udpPort = line.split()
            knownNodes.append((ipAddress, int(udpPort)))
    else:
        knownNodes = None

    node = DistributedTupleSpacePeer(udpPort=int(sys.argv[1]))
    node.joinNetwork(knownNodes)
    twisted.internet.reactor.run()  # @UndefinedVariable



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2011-2013 Codernity (http://codernity.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
import six
import os
import struct
import shutil
import marshal
import io


try:
    from CodernityDB3 import __version__
except ImportError:
    from .__init__ import __version__


class StorageException(Exception):
    pass


class DummyStorage(object):
    """
    Storage mostly used to fake real storage
    """

    def create(self, *args, **kwargs):
        pass

    def open(self, *args, **kwargs):
        pass

    def close(self, *args, **kwargs):
        pass

    def data_from(self, *args, **kwargs):
        pass

    def data_to(self, *args, **kwargs):
        pass

    def save(self, *args, **kwargs):
        return 0, 0

    def insert(self, *args, **kwargs):
        return self.save(*args, **kwargs)

    def update(self, *args, **kwargs):
        return 0, 0

    def get(self, *args, **kwargs):
        return None

    # def compact(self, *args, **kwargs):
    #     pass

    def fsync(self, *args, **kwargs):
        pass

    def flush(self, *args, **kwargs):
        pass


class IU_Storage(object):

    __version__ = __version__

    def __init__(self, db_path, name='main'):
        if isinstance(db_path, bytes):
            db_path = db_path.decode()
        if isinstance(name, bytes):
            name = name.decode()

        self.db_path = db_path
        self.name = name
        self._header_size = 100

    def create(self):
        if os.path.exists(os.path.join(self.db_path, self.name + "_stor")):
            raise IOError("Storage already exists!")
        with io.open(os.path.join(self.db_path, self.name + "_stor"), 'wb') as f:
            if isinstance(self.__version__, six.text_type):
                new_version = self.__version__.encode()
            else:
                new_version = self.__version__
            f.write(struct.pack(b'10s90s', new_version, b'|||||'))
            f.close()
        self._f = io.open(os.path.join(
            self.db_path, self.name + "_stor"), 'r+b', buffering=0)
        self.flush()
        self._f.seek(0, 2)

    def open(self):
        if not os.path.exists(os.path.join(self.db_path, self.name + "_stor")):
            raise IOError("Storage doesn't exists!")
        self._f = io.open(os.path.join(
            self.db_path, self.name + "_stor"), 'r+b', buffering=0)
        self.flush()
        self._f.seek(0, 2)

    def destroy(self):
        os.unlink(os.path.join(self.db_path, self.name + '_stor'))

    def close(self):
        self._f.close()
        # self.flush()
        # self.fsync()

    def data_from(self, data):
        return marshal.loads(data)

    def data_to(self, data):
        return marshal.dumps(data)

    def save(self, data):
        s_data = self.data_to(data)
        self._f.seek(0, 2)
        start = self._f.tell()
        size = len(s_data)
        self._f.write(s_data)
        self.flush()
        # print('storage.save start=%r size=%r' % (start, size))
        return start, size

    def insert(self, data):
        return self.save(data)

    def update(self, data):
        return self.save(data)

    def get(self, start, size, status='c'):
        # print('storage.get start=%r size=%r' % (start, size))
        if status == 'd' or status == b'd':
            return None
        else:
            self._f.seek(start)
            return self.data_from(self._f.read(size))

    def flush(self):
        self._f.flush()

    def fsync(self):
        os.fsync(self._f.fileno())


# classes for public use, done in this way because of
# generation static files with indexes (_index directory)


class Storage(IU_Storage):
    pass



"""

cmd for the addpeers command.

Adds new peers to a running node

Usage:

python3 cmd_addpeers.py ip:port

python3 cmd_addpeers.py ip:port,ip:port,ip:port

if port is omitted, 5658 will be used.
"""

import connections
import json
import socks
import sys

__version__ = '0.0.1'


def add_peers(peers: dict):
    s = socks.socksocket()
    s.settimeout(10)
    s.connect(('127.0.0.1', 5658))
    # Command first
    connections.send(s, 'addpeers')
    # addpeers expects a string, that is a json encoded dict.
    connections.send(s, json.dumps(peers))
    res = connections.receive(s)
    return res


if __name__ == '__main__':
    _, peers_string = sys.argv
    peers = peers_string.split(',')
    peers_dict = {}
    for peer in peers:
        if ':' in peer:
            ip, port = peer.split(':')
        else:
            ip = peer
            port = '5658'
        peers_dict[ip] = port
    res_as_dict = add_peers(peers_dict)
    print('Answer (-1 means busy testing peers):')
    print(json.dumps(res_as_dict, indent=2))



"""
Regnet specific functions and settings
"""

import base64
import math
import os
import sqlite3
import sys
import time
import functools
from Cryptodome.Hash import SHA
from Cryptodome.Signature import PKCS1_v1_5
from hashlib import sha224
from random import getrandbits

import connections
import mempool as mp
import mining_heavy3 as mining

# fixed diff for regnet
REGNET_DIFF = 16

REGNET_PORT = 3030

REGNET_DB = 'static/regmode.db'

REGNET_INDEX = 'static/index_reg.db'

REGNET_PEERS = 'peers_reg.txt'
REGNET_SUGGESTED_PEERS = 'peers_reg.txt'

SQL_INDEX = ['CREATE TABLE aliases (block_height INTEGER, address, alias)', 'CREATE TABLE tokens (block_height INTEGER, timestamp, token, address, recipient, txid, amount INTEGER)']

SQL_LEDGER = [
    'CREATE TABLE misc (block_height INTEGER, difficulty TEXT)',
    'CREATE TABLE transactions (block_height INTEGER, timestamp NUMERIC, address TEXT, recipient TEXT, \
               amount NUMERIC, signature TEXT, public_key TEXT, block_hash TEXT, fee NUMERIC, reward NUMERIC, \
               operation TEXT, openfield TEXT)',
    "INSERT INTO transactions (openfield, operation, reward, fee, block_hash, public_key, signature, \
               amount, recipient, address, timestamp, block_height) \
               VALUES ('genesis', 1, 1, 0, '7a0f384876aca3871adbde8622a87f8b971ede0ed8ee10425e3958a1', \
               '-----BEGIN PUBLIC KEY-----\nMIGfMA0GCSqGSIb3DQEBAQUAA4GNADCBiQKBgQDKvLTbDx85a1ugb/6xMMhVOq6U\n2GeYT8+Iq2z9FwIMR40l2ttGqNK7varNccFLIu8Kn4ogDQs3WSWQCxNkhZh/FqzF\nYYa3/ItPPfzrXqgajwD8q4Zt4Ymjt8+2BkImPjjFNkuTQIz2Iu3yFqOIxLdjMw7n\nUVu9tFPiUkD0VnDPLQIDAQAB\n-----END PUBLIC KEY-----', \
               'DKiWVr+GQHrsEUlu3qEQnsB5rznU4Is7RFLnPmHM1grobiUFHup0kSWiN83gBkNS9LgE57RXUEJvxMKc+9hIAzYE8EwGtO3RsXxkqPTT1v19CguN0iqE4nIM8Bur53/Djs5a1bH/R8EMersemZY1bDJ4jTeeba6yqFxmevGk/gw=', \
               0, '4edadac9093d9326ee4b17f869b14f1a2534f96f9c5d7b48dc9acaed', 'genesis', 1493640955.47, 1);",
    'INSERT INTO misc (difficulty, block_height) VALUES ({},1)'.format(REGNET_DIFF),
]

FILES_TO_REMOVE = [REGNET_DB, REGNET_INDEX]

HASHCOUNT = 10

# Max number of tx to embed per block.
TX_PER_BLOCK = 2

# Do not edit below, it's fed by node.py

ADDRESS = 'This is a fake address placeholder for regtest mode only'
KEY = None
PRIVATE_KEY_READABLE = 'matching priv key'
PUBLIC_KEY_B64ENCODED = 'matching pub key b64'

DIGEST_BLOCK = None

# because of compatibility - huge node refactor wanted.


def sql_trace_callback(log, id, statement):
    line = f'SQL[{id}] {statement}'
    log.warning(line)


def generate_one_block(blockhash, mempool_txs, node, db_handler):
    try:
        if not blockhash:
            node.logger.app_log.warning('Bad blockhash')
            return
        diff_hex = math.floor((REGNET_DIFF/8) - 1)
        mining_condition = blockhash[0:diff_hex]
        while True:
            try_arr = [('%0x' % getrandbits(32)) for i in range(HASHCOUNT)]
            i = 0
            for i in range(100):
                i += 1
                seed = ('%0x' % getrandbits(128 - 32))
                prefix = ADDRESS + seed
                # print("node heavy", node.heavy)
                if node.heavy:
                    possibles = [nonce for nonce in try_arr if mining_condition in (mining.anneal3(mining.MMAP, int.from_bytes(sha224((prefix + nonce + blockhash).encode('utf-8')).digest(), 'big')))]
                else:
                    possibles = [nonce for nonce in try_arr if mining_condition in (mining.anneal3_regnet(mining.MMAP, int.from_bytes(sha224((prefix + nonce + blockhash).encode('utf-8')).digest(), 'big')))]
                if possibles:
                    nonce = seed + possibles[0]
                    node.logger.app_log.warning('Generate got a block in {} tries len {}'.format(i, len(possibles)))
                    # assemble block with mp data
                    txs = []
                    for i in range(TX_PER_BLOCK):
                        if not len(mempool_txs):
                            break
                        txs.append(mempool_txs.pop(0))
                    block_send = []
                    removal_signature = []
                    for mpdata in txs:
                        transaction = (str(mpdata[0]), str(mpdata[1][:56]), str(mpdata[2][:56]), '%.8f' % float(mpdata[3]), str(mpdata[4]), str(mpdata[5]), str(mpdata[6]), str(mpdata[7]))  # create tuple
                        # node.logger.app_log.warning transaction
                        block_send.append(transaction)  # append tuple to list for each run
                        removal_signature.append(str(mpdata[4]))  # for removal after successful mining
                    # claim reward
                    block_timestamp = '%.2f' % time.time()
                    transaction_reward = (str(block_timestamp), str(ADDRESS[:56]), str(ADDRESS[:56]), '%.8f' % float(0), '0', str(nonce))  # only this part is signed!
                    # node.logger.app_log.warning transaction_reward

                    hash = SHA.new(str(transaction_reward).encode('utf-8'))
                    signer = PKCS1_v1_5.new(KEY)
                    signature = signer.sign(hash)
                    signature_enc = base64.b64encode(signature)

                    if signer.verify(hash, signature):
                        node.logger.app_log.warning('Signature valid')
                        block_send.append((str(block_timestamp), str(ADDRESS[:56]), str(ADDRESS[:56]), '%.8f' % float(0), str(signature_enc.decode('utf-8')), str(PUBLIC_KEY_B64ENCODED.decode('utf-8')), '0', str(nonce)))  # mining reward tx
                        # node.logger.app_log.warning('Block to send: {}'.format(block_send))
                    # calc hash

                    new_hash = DIGEST_BLOCK(node, [block_send], None, 'regtest', db_handler)
                    # post block to self or better, send to db to make sure it is. when we add the next one?
                    # use a link to the block digest function
                    # embed at mot TX_PER_BLOCK txs from the mp
                    return new_hash

    except Exception as e:
        node.logger.app_log.warning(e)
        exc_type, exc_obj, exc_tb = sys.exc_info()
        fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
        node.logger.app_log.warning(exc_type, fname, exc_tb.tb_lineno)


def command(sdef, data, blockhash, node, db_handler):
    try:
        node.logger.app_log.warning('Regnet got command {}'.format(data))
        if data == 'regtest_generate':
            how_many = int(connections.receive(sdef))
            node.logger.app_log.warning('regtest_generate {} {}'.format(how_many, blockhash))
            mempool_txs = mp.MEMPOOL.fetchall(mp.SQL_SELECT_TX_TO_SEND)
            for i in range(how_many):
                blockhash = generate_one_block(blockhash, mempool_txs, node, db_handler)
            connections.send(sdef, 'OK')
    except Exception as e:
        node.logger.app_log.warning(e)
        exc_type, exc_obj, exc_tb = sys.exc_info()
        fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
        node.logger.app_log.warning(exc_type, fname, exc_tb.tb_lineno)


def init(app_log, trace_db_calls=False):
    # Empty peers
    with open(REGNET_PEERS, 'w') as f:
        f.write('{}')
    with open(REGNET_SUGGESTED_PEERS, 'w') as f:
        f.write('{}')
    # empty files
    for remove_me in FILES_TO_REMOVE:
        if os.path.exists(remove_me):
            os.remove(remove_me)
    # create empty index db
    with sqlite3.connect(REGNET_DB) as source_db:
        if trace_db_calls:
            source_db.set_trace_callback(functools.partial(sql_trace_callback, app_log, 'REGNET-INIT'))
        for request in SQL_LEDGER:
            source_db.execute(request)
        source_db.commit()
    # create empty reg db
    with sqlite3.connect(REGNET_INDEX) as source_db:
        if trace_db_calls:
            source_db.set_trace_callback(functools.partial(sql_trace_callback, app_log, 'REGNET-INIT-INDEX'))
        for request in SQL_INDEX:
            source_db.execute(request)
        source_db.commit()

    # Here, we do not have the keys info yet.



#!/usr/bin/env python
# encoding.py
#
# Copyright (C) 2007-2008 Francois Aucamp, Meraka Institute, CSIR
# See AUTHORS for all authors and contact information.
#
# License: GNU Lesser General Public License, version 3 or later; see COPYING
#          included in this archive for details.
#
# This library is free software, distributed under the terms of
# the GNU Lesser General Public License Version 3, or any later version.
# See the COPYING file included in this archive
#
# The docstrings in this module contain epytext markup; API documentation
# may be created by processing this file with epydoc: http://epydoc.sf.net

from __future__ import absolute_import
from __future__ import unicode_literals
import six
import codecs
import sys

if sys.version_info[0] == 3:
    text_type = str
    binary_type = bytes
else:
    text_type = unicode  # @UndefinedVariable
    binary_type = str

_Debug = False


def is_text(s):
    """
    Return `True` if `s` is a text value:
    + `unicode()` in Python2
    + `str()` in Python3
    """
    return isinstance(s, text_type)


def is_bin(s):
    """
    Return `True` if `s` is a binary value:
    + `str()` in Python2
    + `bytes()` in Python3
    """
    return isinstance(s, binary_type)


def is_string(s):
    """
    Return `True` if `s` is text or binary type (not integer, class, list, etc...)
    """
    return is_text(s) or is_bin(s)


def to_text(s, encoding='utf-8', errors='strict'):
    """
    If ``s`` is binary type - decode it to unicode - "text" type in Python3 terms.
    If ``s`` is not binary and not text calls `str(s)` to build text representation.
    """
    if s is None:
        return s
    if not is_string(s):
        s = text_type(s)
    if is_text(s):
        return s
    return s.decode(encoding=encoding, errors=errors)


def to_bin(s, encoding='utf-8', errors='strict'):
    """
    If ``s`` is unicode ("text" type in Python3 terms) - encode it to utf-8, otherwise return ``s``.
    If ``s`` is not binary and not text calls `str(s)` to build text representation,
    then encode to binary and return.
    """
    if s is None:
        return s
    if not is_string(s):
        s = text_type(s)
    if is_bin(s):
        return s
    return s.encode(encoding=encoding, errors=errors)


def encode_hex(value, as_string=True, encoding='utf-8'):
    hex_value = codecs.encode(value, 'hex')
    if as_string:
        return hex_value.decode('utf-8')
    return hex_value


def decode_hex(value, as_string=True, encoding='utf-8'):
    orig_value = codecs.decode(value, 'hex')
    if as_string:
        return orig_value.decode('utf-8')
    return orig_value


class Encoding(object):
    """
    Interface for RPC message encoders/decoders.

    All encoding implementations used with this library should inherit
    and implement this.
    """
    def encode(self, data):
        """
        Encode the specified data.

        @param data: The data to encode
                     This method has to support encoding of the following
                     types: C{str}, C{int} and C{long}
                     Any additional data types may be supported as long as the
                     implementing class's C{decode()} method can successfully
                     decode them.

        @return: The encoded data
        @rtype: str
        """

    def decode(self, data):
        """
        Decode the specified data string.

        @param data: The data (byte string) to decode.
        @type data: str

        @return: The decoded data (in its correct type)
        """


class Bencode(Encoding):
    """
    Implementation of a Bencode-based algorithm (Bencode is the encoding
    algorithm used by Bittorrent).

    @note: This algorithm differs from the "official" Bencode algorithm in
           that it can encode/decode floating point values in addition to
           integers.
    """
    def encode(self, data, encoding='utf-8'):
        """
        Encoder implementation of the Bencode algorithm.

        @param data: The data to encode
        @type data: int, long, tuple, list, dict or str

        @return: The encoded data
        @rtype: str
        """
        def _e():
            if data is None:
                return b'i0e'  # return 0
            elif type(data) in six.integer_types:
                return b'i%de' % data
            elif isinstance(data, six.text_type):
                return b'%d:%s' % (len(data), data.encode(encoding=encoding))
            elif isinstance(data, six.binary_type):
                return b'%d:%s' % (len(data), data)
            elif type(data) in (list, tuple):
                encodedListItems = b''
                for item in data:
                    encodedListItems += self.encode(item)
                return b'l%se' % encodedListItems
            elif isinstance(data, dict):
                encodedDictItems = b''
                _d = {}
                for k, v in data.items():
                    e_key = self.encode(k)
                    _d[e_key] = v
                # _d = {(k.encode() if k and isinstance(k, six.text_type) else k) : v for k, v in data.items()}
                keys = sorted(_d.keys())
                for e_key in keys:
                    e_data = self.encode(_d[e_key])
                    encodedDictItems += e_key
                    encodedDictItems += e_data
                return b'd%se' % encodedDictItems
            elif isinstance(data, float):
                # This (float data type) is a non-standard extension to the original Bencode algorithm
                return b'f%fe' % data
            else:
                raise TypeError("Cannot bencode '%s' object" % type(data))

        try:
            ret = _e()
            if _Debug:
                print('[DHT ENCODING]         encode  %r  into  %d bytes' % (
                    type(data),
                    len(ret),
                ))
            return ret
        except Exception as exc:
            if _Debug:
                print('[DHT ENCODING]         encode failed with: %r' % exc)

    def decode(self, data, encoding=None):
        """
        Decoder implementation of the Bencode algorithm.

        @param data: The encoded data
        @type data: str

        @note: This is a convenience wrapper for the recursive decoding
               algorithm, C{_decodeRecursive}

        @return: The decoded data, as a native Python type
        @rtype:  int, list, dict or str
        """
        try:
            ret, endPos = self._decodeRecursive(data, encoding=encoding)
            if _Debug:
                print('[DHT ENCODING]         decode %r  endPos=%d' % (
                    type(ret),
                    endPos,
                ))
            return ret
        except Exception as exc:
            if _Debug:
                print('[DHT ENCODING]         decode failed with: %r' % exc)

    @staticmethod
    def _decodeRecursive(data, startIndex=0, encoding=None):
        """
        Actual implementation of the recursive Bencode algorithm.

        Do not call this; use C{decode()} instead
        """
        if data[startIndex:startIndex + 1] == b'i':
            endPos = data[startIndex:].find(b'e') + startIndex
            return (int(to_text(data[startIndex + 1:endPos]) or '0'), endPos + 1)
        elif data[startIndex:startIndex + 1] == b'l':
            startIndex += 1
            decodedList = []
            while data[startIndex:startIndex + 1] != b'e':
                listData, startIndex = Bencode._decodeRecursive(data, startIndex, encoding=encoding)
                decodedList.append(listData)
            return (decodedList, startIndex + 1)
        elif data[startIndex:startIndex + 1] == b'd':
            startIndex += 1
            decodedDict = {}
            while data[startIndex:startIndex + 1] != b'e':
                key, startIndex = Bencode._decodeRecursive(data, startIndex, encoding=encoding)
                value, startIndex = Bencode._decodeRecursive(data, startIndex, encoding=encoding)
                decodedDict[key] = value
            return (decodedDict, startIndex)
        elif data[startIndex:startIndex + 1] == b'f':
            # This (float data type) is a non-standard extension to the original Bencode algorithm
            endPos = data[startIndex:].find(b'e') + startIndex
            return (float(to_text(data[startIndex + 1:endPos]) or '0'), endPos + 1)
        else:
            splitPos = data[startIndex:].find(b':') + startIndex
            length = int(to_text(data[startIndex:splitPos]) or '0')
            startIndex = splitPos + 1
            endPos = startIndex + length
            byts = data[startIndex:endPos]
            if encoding:
                byts = byts.decode(encoding=encoding)
            return (byts, endPos)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2011-2013 Codernity (http://codernity.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
import six

from bitdust_forks.CodernityDB3.env import cdb_environment
from bitdust_forks.CodernityDB3.database import PreconditionsException, RevConflict, Database
# from database import Database

from collections import defaultdict
from functools import wraps
from types import MethodType


class th_safe_gen:

    def __init__(self, name, gen, l=None):
        self.lock = l
        self.__gen = gen
        self.name = name

    def __iter__(self):
        return self

    def __next__(self):
        with self.lock:
            return next(self.__gen)

    @staticmethod
    def wrapper(method, index_name, meth_name, l=None):
        @wraps(method)
        def _inner(*args, **kwargs):
            res = method(*args, **kwargs)
            return th_safe_gen(index_name + "_" + meth_name, res, l)
        return _inner


def safe_wrapper(method, lock):
    @wraps(method)
    def _inner(*args, **kwargs):
        with lock:
            return method(*args, **kwargs)
    return _inner


class SafeDatabase(Database):

    def __init__(self, path, *args, **kwargs):
        super(SafeDatabase, self).__init__(path, *args, **kwargs)
        self.indexes_locks = defaultdict(
            lambda: cdb_environment['rlock_obj']())
        self.close_open_lock = cdb_environment['rlock_obj']()
        self.main_lock = cdb_environment['rlock_obj']()
        self.id_revs = {}

    def __patch_index_gens(self, name):
        ind = self.indexes_names[name]
        for c in ('all', 'get_many'):
            m = getattr(ind, c)
            if getattr(ind, c + "_orig", None):
                return
            m_fixed = th_safe_gen.wrapper(m, name, c, self.indexes_locks[name])
            setattr(ind, c, m_fixed)
            setattr(ind, c + '_orig', m)

    def __patch_index_methods(self, name):
        ind = self.indexes_names[name]
        lock = self.indexes_locks[name]
        for curr in dir(ind):
            meth = getattr(ind, curr)
            if not curr.startswith('_') and isinstance(meth, MethodType):
                setattr(ind, curr, safe_wrapper(meth, lock))
        stor = ind.storage
        for curr in dir(stor):
            meth = getattr(stor, curr)
            if not curr.startswith('_') and isinstance(meth, MethodType):
                setattr(stor, curr, safe_wrapper(meth, lock))

    def __patch_index(self, name):
        self.__patch_index_methods(name)
        self.__patch_index_gens(name)

    def initialize(self, *args, **kwargs):
        with self.close_open_lock:
            self.close_open_lock.acquire()
            res = super(SafeDatabase, self).initialize(*args, **kwargs)
            for name in self.indexes_names.keys():
                self.indexes_locks[name] = cdb_environment['rlock_obj']()
            return res

    def open(self, *args, **kwargs):
        with self.close_open_lock:
            res = super(SafeDatabase, self).open(*args, **kwargs)
            for name in self.indexes_names.keys():
                self.indexes_locks[name] = cdb_environment['rlock_obj']()
                self.__patch_index(name)
            return res

    def create(self, *args, **kwargs):
        with self.close_open_lock:
            res = super(SafeDatabase, self).create(*args, **kwargs)
            for name in self.indexes_names.keys():
                self.indexes_locks[name] = cdb_environment['rlock_obj']()
                self.__patch_index(name)
            return res

    def close(self):
        with self.close_open_lock:
            return super(SafeDatabase, self).close()

    def destroy(self):
        with self.close_open_lock:
            return super(SafeDatabase, self).destroy()

    def add_index(self, *args, **kwargs):
        with self.main_lock:
            res = super(SafeDatabase, self).add_index(*args, **kwargs)
            if self.opened:
                self.indexes_locks[res] = cdb_environment['rlock_obj']()
                self.__patch_index(res)
            return res

    def _single_update_index(self, index, data, db_data, doc_id):
        with self.indexes_locks[index.name]:
            super(SafeDatabase, self)._single_update_index(
                index, data, db_data, doc_id)

    def _single_delete_index(self, index, data, doc_id, old_data):
        with self.indexes_locks[index.name]:
            super(SafeDatabase, self)._single_delete_index(
                index, data, doc_id, old_data)

    def edit_index(self, *args, **kwargs):
        with self.main_lock:
            res = super(SafeDatabase, self).edit_index(*args, **kwargs)
            if self.opened:
                self.indexes_locks[res] = cdb_environment['rlock_obj']()
                self.__patch_index(res)
            return res

    def set_indexes(self, *args, **kwargs):
        try:
            self.main_lock.acquire()
            super(SafeDatabase, self).set_indexes(*args, **kwargs)
        finally:
            self.main_lock.release()

    def reindex_index(self, index, *args, **kwargs):
        if isinstance(index, six.string_types):
            if not index in self.indexes_names:
                raise PreconditionsException("No index named %s" % index)
            index = self.indexes_names[index]
        key = index.name + "reind"
        self.main_lock.acquire()
        if key in self.indexes_locks:
            lock = self.indexes_locks[index.name + "reind"]
        else:
            self.indexes_locks[index.name +
                               "reind"] = cdb_environment['rlock_obj']()
            lock = self.indexes_locks[index.name + "reind"]
        self.main_lock.release()
        try:
            lock.acquire()
            super(SafeDatabase, self).reindex_index(
                index, *args, **kwargs)
        finally:
            lock.release()

    def flush(self):
        try:
            self.main_lock.acquire()
            super(SafeDatabase, self).flush()
        finally:
            self.main_lock.release()

    def fsync(self):
        try:
            self.main_lock.acquire()
            super(SafeDatabase, self).fsync()
        finally:
            self.main_lock.release()

    def _update_id_index(self, _rev, data):
        with self.indexes_locks['id']:
            return super(SafeDatabase, self)._update_id_index(_rev, data)

    def _delete_id_index(self, _id, _rev, data):
        with self.indexes_locks['id']:
            return super(SafeDatabase, self)._delete_id_index(_id, _rev, data)

    def _update_indexes(self, _rev, data):
        _id, new_rev, db_data = self._update_id_index(_rev, data)
        with self.main_lock:
            self.id_revs[_id] = new_rev
        for index in self.indexes[1:]:
            with self.main_lock:
                curr_rev = self.id_revs.get(_id)  # get last _id, _rev
                if curr_rev != new_rev:
                    break  # new update on the way stop current
            self._single_update_index(index, data, db_data, _id)
        with self.main_lock:
            if self.id_revs[_id] == new_rev:
                del self.id_revs[_id]
        return _id, new_rev

    def _delete_indexes(self, _id, _rev, data):
        old_data = self.get('id', _id)
        if old_data['_rev'] != _rev:
            raise RevConflict()
        with self.main_lock:
            self.id_revs[_id] = _rev
        for index in self.indexes[1:]:
            self._single_delete_index(index, data, _id, old_data)
        self._delete_id_index(_id, _rev, data)
        with self.main_lock:
            if self.id_revs[_id] == _rev:
                del self.id_revs[_id]



#!/usr/bin/env python
# supplier_contract_executor.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (supplier_contract_executor.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
"""
.. module:: supplier_contract_executor
.. role:: red

BitDust supplier_contract_executor() Automat

EVENTS:
    * :red:`chain-1-state`
    * :red:`chain-2-state`
    * :red:`chain-closed`
    * :red:`chain-empty`
    * :red:`coin-failed`
    * :red:`coin-published`
    * :red:`coin-sent`
    * :red:`contract-signed`
    * :red:`init`
    * :red:`payment-timeout`
    * :red:`query-failed`
    * :red:`recheck`
    * :red:`shutdown`
    * :red:`time-to-charge`
    * :red:`timer-1min`
    * :red:`timer-30sec`
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 6

#------------------------------------------------------------------------------

if __name__ == '__main__':
    import sys
    import os.path as _p
    sys.path.insert(0, _p.abspath(_p.join(_p.dirname(_p.abspath(sys.argv[0])), '..')))

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.automats import automat

from bitdust.lib import nameurl

from bitdust.storage import accounting

from bitdust.userid import my_id

from bitdust.p2p import commands
from bitdust.p2p import p2p_service

from bitdust.coins import coins_io
from bitdust.coins import contract_chain_node

#------------------------------------------------------------------------------

_ActiveSupplierContracts = dict()  # provides SupplierContractExecutor object by customer idurl

#------------------------------------------------------------------------------


def all_contracts():
    global _ActiveSupplierContracts
    return _ActiveSupplierContracts


def init_contract(customer_idurl):
    global _ActiveSupplierContracts
    if customer_idurl in _ActiveSupplierContracts:
        lg.warn('contract with customer %s already started' % customer_idurl)
        return _ActiveSupplierContracts[customer_idurl]
    cce = SupplierContractExecutor(customer_idurl)
    cce.automat('init')
    _ActiveSupplierContracts[customer_idurl] = cce
    return _ActiveSupplierContracts[customer_idurl]


def shutdown_contract(customer_idurl):
    global _ActiveSupplierContracts
    if customer_idurl not in _ActiveSupplierContracts:
        lg.warn('no contract started for given customer')
        return False
    _ActiveSupplierContracts[customer_idurl].automat('shutdown')
    _ActiveSupplierContracts.pop(customer_idurl)
    return True


def get_contract(customer_idurl):
    global _ActiveSupplierContracts
    if customer_idurl not in _ActiveSupplierContracts:
        return None
    return _ActiveSupplierContracts[customer_idurl]


def recheck_contract(customer_idurl):
    contract_executor = get_contract(customer_idurl)
    if contract_executor is None:
        contract_executor = init_contract(customer_idurl)
    contract_executor.automat('recheck')


#------------------------------------------------------------------------------


class SupplierContractExecutor(automat.Automat):
    """
    This class implements all the functionality of the ``supplier_contract_executor()`` state machine.
    """

    timers = {
        'timer-1min': (60, ['READ_CHAIN?']),
        'timer-30sec': (30.0, ['CUSTOMER_SIGN?']),
    }

    def __init__(self, customer_idurl):
        self.customer_idurl = customer_idurl
        name = 'supplier_executor_%s' % nameurl.GetName(self.customer_idurl)
        automat.Automat.__init__(self, name, 'AT_STARTUP', _DebugLevel, _Debug)

    def init(self):
        """
        Method to initialize additional variables and flags
        at creation phase of supplier_contract_executor() machine.
        """
        self.current_duration = 60*60  # TODO: read from settings

    def A(self, event, *args, **kwargs):
        """
        The state machine code, generated using `visio2python <https://bitdust.io/visio2python/>`_ tool.
        """
        #---AT_STARTUP---
        if self.state == 'AT_STARTUP':
            if event == 'init':
                self.state = 'READ_CHAIN?'
                self.doRequestCoins(*args, **kwargs)
        #---MY_COIN!---
        elif self.state == 'MY_COIN!':
            if event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'coin-failed':
                self.state = 'UNCLEAR'
            elif event == 'coin-sent':
                self.state = 'CUSTOMER_COIN?'
        #---CUSTOMER_COIN?---
        elif self.state == 'CUSTOMER_COIN?':
            if event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'coin-published':
                self.state = 'ACTIVE'
                self.doSchedulePayment(*args, **kwargs)
            elif event == 'payment-timeout':
                self.state = 'FINISHED'
                self.doRemoveCustomer(*args, **kwargs)
        #---ACTIVE---
        elif self.state == 'ACTIVE':
            if event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'time-to-charge':
                self.state = 'MY_COIN!'
                self.doSendNextCoin(*args, **kwargs)
        #---CLOSED---
        elif self.state == 'CLOSED':
            pass
        #---FINISHED---
        elif self.state == 'FINISHED':
            if event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
        #---UNCLEAR---
        elif self.state == 'UNCLEAR':
            if event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'recheck':
                self.state = 'READ_CHAIN?'
                self.doRequestCoins(*args, **kwargs)
        #---READ_CHAIN?---
        elif self.state == 'READ_CHAIN?':
            if event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
            elif event == 'chain-2-state':
                self.state = 'ACTIVE'
                self.doSchedulePayment(*args, **kwargs)
            elif event == 'timer-1min' or event == 'query-failed':
                self.state = 'UNCLEAR'
            elif event == 'chain-empty':
                self.state = 'CUSTOMER_SIGN?'
                self.doRequestCustomerSignature(*args, **kwargs)
            elif event == 'chain-1-state':
                self.state = 'CUSTOMER_COIN?'
            elif event == 'chain-closed':
                self.state = 'FINISHED'
        #---CUSTOMER_SIGN?---
        elif self.state == 'CUSTOMER_SIGN?':
            if event == 'timer-30sec':
                self.state = 'UNCLEAR'
            elif event == 'contract-signed':
                self.state = 'MY_COIN!'
                self.doSendFirstCoin(*args, **kwargs)
            elif event == 'shutdown':
                self.state = 'CLOSED'
                self.doDestroyMe(*args, **kwargs)
        return None

    def isCustomerCoinExist(self, coins):
        """
        Condition method.
        """
        if not coins:
            return False
        # TODO:
        return True

    def isChainClosed(self, coins):
        """
        Condition method.
        """
        if not coins:
            return False
        # TODO:
        return True

    def isMyCoinExist(self, coins):
        """
        Condition method.
        """
        if not coins:
            return False
        # TODO:
        return True

    def doRequestCoins(self, *args, **kwargs):
        """
        Action method.
        """
        contract_chain_node.get_coins_by_chain(
            chain='supplier_customer',
            provider_idurl=my_id.getIDURL(),
            consumer_idurl=self.customer_idurl,
        ).addCallbacks(
            self._on_query_result,
            self._on_query_failed,
        )

    def doRequestCustomerSignature(self, *args, **kwargs):
        """
        Action method.
        """
        bytes_allocated = accounting.get_customer_quota(self.customer_idurl)
        assert bytes_allocated is not None
        assert bytes_allocated > 0
        coin_json = coins_io.storage_contract_open(
            self.customer_idurl,
            self.current_duration,
            bytes_allocated,
        )
        coin_json_sined = coins_io.add_signature(coin_json, 'creator')
        p2p_service.SendCoin(
            self.customer_idurl,
            [
                coin_json_sined,
            ],
            callbacks={
                commands.Ack(): self._on_signature_ack,
                commands.Fail(): self._on_signature_fail,
            },
        )

    def doSendFirstCoin(self, *args, **kwargs):
        """
        Action method.
        """
        coin_json = args[0]
        contract_chain_node.send_to_miner([
            coin_json,
        ], ).addCallbacks(
            self._on_coin_mined,
            self._on_coin_failed,
        )

    def doSendNextCoin(self, *args, **kwargs):
        """
        Action method.
        """

    def doDestroyMe(self, *args, **kwargs):
        """
        Remove all references to the state machine object to destroy it.
        """
        self.destroy()

    def doRemoveCustomer(self, *args, **kwargs):
        """
        Action method.
        """

    def doSchedulePayment(self, *args, **kwargs):
        """
        Action method.
        """

    def _on_query_result(self, coins):
        if _Debug:
            lg.out(_DebugLevel, 'supplier_contract_executor._on_coins_received %s' % coins)
        if coins is None:
            self.automat('query-failed', [])
            return
        if len(coins) == 0:
            self.automat('chain-empty', [])
            return
        if len(coins) % 2 == 1:
            self.automat('chain-1-state', coins)
            return
        if len(coins) % 2 == 0:
            self.automat('chain-2-state', coins)
            return
        self.automat('query-failed', coins)

    def _on_query_failed(self, fails):
        if _Debug:
            lg.warn(str(fails))
        self.automat('query-failed', fails)

    def _on_signature_ack(self, response, info):
        pass

    def _on_signature_fail(self, response, info):
        pass



#!/usr/bin/env python
# global_state.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (global_state.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
#
"""
.. module:: global_state.

This module is to keep track of changing states of State Machines.
It also remember the current ``global`` state of the program - this a stats of a several most important automats.
"""

from __future__ import absolute_import

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.automats import automat

#------------------------------------------------------------------------------

_StatesDict = {
    'init at startup': 'beginning',
    'init local': 'local settings initialization',
    'init contacts': 'contacts initialization',
    'init connection': 'initializing connections',
    'init modules': 'starting modules',
    'init install': 'preparing install section',
    'network at startup': 'starting connection',
    'network stun': 'detecting external IP address',
    'network upnp': 'configuring UPnP',
    'network connected': 'internet connection is fine',
    'network disconnected': 'internet connection is not working',
    'network network?': 'checking network interfaces',
    'network google?': 'is www.google.com available?',
    'p2p at startup': 'initial peer-to-peer state',
    'p2p transports': 'starting network transports',
    'p2p propagate': 'propagate my identity',
    'p2p incomming?': 'waiting response from others',
    'p2p connected': 'ready',
    'p2p disconnected': 'starting disconnected',
}

_GlobalState = 'AT_STARTUP'
_GlobalStateNotifyFunc = None

#------------------------------------------------------------------------------


def set_global_state(st):
    """
    This method is called from State Machines when ``state`` is changed:
    global_state.set_global_state('P2P ' + newstate) So ``st`` is a string
    like: 'P2P CONNECTED'.

    ``_GlobalStateNotifyFunc`` can be used to keep track of changing
    program states.
    """
    global _GlobalState
    global _GlobalStateNotifyFunc
    oldstate = _GlobalState
    _GlobalState = st
    # lg.out(6, (' ' * 40) + '{%s}->{%s}' % (oldstate, _GlobalState))
    if _GlobalStateNotifyFunc is not None and oldstate != _GlobalState:
        try:
            _GlobalStateNotifyFunc(_GlobalState)
        except:
            lg.exc()


def get_global_state():
    """
    Return the current ``global state``, for example: P2P CONNECTED.
    """
    global _GlobalState
    # lg.out(6, 'global_state.get_global_state return [%s]' % _GlobalState)
    return _GlobalState


def get_global_state_label():
    """
    Return a label describing current global state, for example: 'checking
    network interfaces'.
    """
    global _GlobalState
    global _StatesDict
    return _StatesDict.get(_GlobalState.replace('_', ' ').lower(), '')


def SetGlobalStateNotifyFunc(f):
    """
    Set callback to catch a global state changed event.
    """
    global _GlobalStateNotifyFunc
    _GlobalStateNotifyFunc = f


def SetSingleStateNotifyFunc(f):
    """
    Set callback to catch state change of any automat.
    """
    automat.SetStateChangedCallback(f)


#------------------------------------------------------------------------------



# optihash.py v 0.30 to be used with Python3.5 or better
# Optimized CPU-miner for Optipoolware based pool mining only
# Copyright Hclivess, Primedigger, Maccaspacca, SylvainDeaure 2017
# .

import time, socks, connections, sys, math
from multiprocessing import Process, freeze_support, Queue
from random import getrandbits
from hashlib import sha224

import mining_heavy3 as mining
#import annealing

__version__ = '0.3.1'

# load config
lines = [line.rstrip('\n') for line in open('miner.txt')]
for line in lines:
    if 'port=' in line:
        port = line.split('=')[1]
    if 'mining_ip=' in line:
        mining_ip_conf = line.split('=')[1]
    if 'mining_threads=' in line:
        mining_threads_conf = line.strip('mining_threads=')
    if 'tor=' in line:
        tor_conf = int(line.strip('tor='))
    if 'miner_address=' in line:
        self_address = line.split('=')[1]
    if 'nonce_time=' in line:
        nonce_time = int(line.split('=')[1])
    if 'miner_name=' in line:
        mname = line.split('=')[1]
    if 'hashcount=' in line:
        hashcount = int(line.split('=')[1])

# load config

bin_format_dict = dict((x, format(ord(x), '8b').replace(' ', '0')) for x in '0123456789abcdef')

try:
    how_much_coins = int(sys.argv[1])
except:
    how_much_coins = 100000000


def bin_convert(string):
    return ''.join(bin_format_dict[x] for x in string)


def bin_convert_orig(string):
    return ''.join(format(ord(x), '8b').replace(' ', '0') for x in string)


def diffme(pool_address, nonce, db_block_hash):
    # minimum possible diff
    diff = 60
    # will return 0 for diff < 60
    diff_result = 0
    mining_hash = bin_convert(sha224((pool_address + nonce + db_block_hash).encode('utf-8')).hexdigest())
    mining_condition = bin_convert(db_block_hash)
    while mining_condition[:diff] in mining_hash:
        diff_result = diff
        diff += 1
    return diff_result


def miner(q, pool_address, db_block_hash, diff, mining_condition, netdiff, hq, thr, dh):
    global how_much_coins
    mined_coins = 0
    process_mmap = False
    if not mining.RND_LEN:
        mining.mining_open()
        process_mmap = True
    try:
        tries = 0
        try_arr = [('%0x' % getrandbits(32)) for i in range(nonce_time*hashcount)]
        address = pool_address
        timeout = time.time() + nonce_time
        # print(pool_address)
        while time.time() < timeout:
            if mined_coins >= how_much_coins:
                break
            try:
                t1 = time.time()
                tries = tries + 1
                # generate the "address" of a random backyard that we will sample in this try
                seed = ('%0x' % getrandbits(128 - 32))
                # this part won't change, so concat once only
                prefix = pool_address + seed
                # This is where the actual hashing takes place
                # possibles = [nonce for nonce in try_arr if mining_condition in (sha224((prefix + nonce + db_block_hash).encode("utf-8")).hexdigest())]
                possibles = [nonce for nonce in try_arr if mining_condition in (mining.anneal3(mining.MMAP, int.from_bytes(sha224((prefix + nonce + db_block_hash).encode('utf-8')).digest(), 'big')))]
                # hash rate calculation
                try:
                    t2 = time.time()
                    h1 = int(((nonce_time*hashcount)/(t2 - t1))/1000)
                except Exception as e:
                    h1 = 1
                if possibles:
                    # print(possibles)
                    for nonce in possibles:
                        # add the seed back to get a full 128 bits nonce
                        nonce = seed + nonce
                        # xdiffx = diffme(str(address[:56]),str(nonce),db_block_hash)
                        xdiffx = mining.diffme_heavy3(address, nonce, db_block_hash)
                        if xdiffx < diff:
                            pass
                        else:
                            print('Thread {} solved work with difficulty {} in {} cycles - YAY!'.format(q, xdiffx, tries))
                            wname = '{}{}'.format(mname, str(q))
                            print('{} running at {} kh/s'.format(wname, str(h1)))
                            block_send = []
                            del block_send[:]  # empty
                            block_timestamp = '%.2f' % time.time()
                            block_send.append((block_timestamp, nonce, db_block_hash, netdiff, xdiffx, dh, mname, thr, str(q)))
                            print('Sending solution: {}'.format(block_send))
                            tries = 0
                            # submit mined nonce to pool
                            try:
                                s1 = socks.socksocket()
                                if tor_conf == 1:
                                    s1.setproxy(socks.PROXY_TYPE_SOCKS5, '127.0.0.1', 9050)
                                s1.connect((mining_ip_conf, int(port)))  # connect to pool
                                print('Miner: connected to pool, proceeding to submit solution')
                                connections.send(s1, 'block', 10)
                                connections.send(s1, self_address, 10)
                                connections.send(s1, block_send, 10)
                                time.sleep(0.2)
                                s1.close()
                                mined_coins += 1
                                print('Miner: solution submitted to pool', mined_coins, how_much_coins)
                                if mined_coins >= how_much_coins:
                                    break

                            except Exception as e:
                                print('Miner: Could not submit solution to pool')
            except Exception as e:
                print(e)
                time.sleep(0.1)
                raise
        hq.put(str(h1) + '_' + str(mined_coins))
    finally:
        if process_mmap:
            mining.mining_close()


def runit():
    global how_much_coins
    totoally_mined_coins = 0
    connected = 0
    dh = 0
    hq = Queue()

    while True:
        try:

            s = socks.socksocket()
            if tor_conf == 1:
                s.setproxy(socks.PROXY_TYPE_SOCKS5, '127.0.0.1', 9050)
            s.connect((mining_ip_conf, int(port)))  # connect to pool
            connections.send(s, 'getwork', 10)
            work_pack = connections.receive(s, 10)
            db_block_hash = (work_pack[-1][0])
            diff = int((work_pack[-1][1]))
            paddress = (work_pack[-1][2])
            netdiff = int((work_pack[-1][3]))
            s.close()

            diff_hex = math.floor((diff/8) - 1)
            mining_condition = db_block_hash[0:diff_hex]

            instances = range(int(mining_threads_conf))
            thr = int(mining_threads_conf)

            for q in instances:
                p = Process(target=miner, args=(str(q + 1), paddress, db_block_hash, diff, mining_condition, netdiff, hq, thr, dh))
                p.daemon = True
                p.start()
            print('{} miners searching for solutions at difficulty {} and condition {}'.format(mining_threads_conf, str(diff), str(mining_condition)))

            time.sleep(nonce_time)

            for q in instances:
                p.join()
                p.terminate()

            raw_results = [hq.get() for q in instances]
            results = [int(r.split('_')[0]) for r in raw_results]
            totoally_mined_coins += sum([int(r.split('_')[1]) for r in raw_results])
            dh = sum(results)
            print('Current total hash rate is {} kh/s : in total {} coins mined '.format(str(dh), totoally_mined_coins))
            if totoally_mined_coins >= how_much_coins*len(instances):
                break

        except Exception as e:
            print(e)
            print('Miner: Unable to connect to pool check your connection or IP settings.')
            time.sleep(1)


if __name__ == '__main__':
    freeze_support()  # must be this line, don't move ahead

    mining.mining_open()
    try:
        runit()
    finally:
        mining.mining_close()



#!/usr/bin/python
# http_node.py
#
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (http_node.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
.. module:: http_node.

"""

#------------------------------------------------------------------------------

from __future__ import absolute_import
from __future__ import print_function

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 8

#------------------------------------------------------------------------------

import os
import sys
import time
import base64

#------------------------------------------------------------------------------

try:
    from twisted.internet import reactor  # @UnresolvedImport
except:
    sys.exit('Error initializing twisted.internet.reactor in tcp_node.py')

from twisted.internet.defer import Deferred
from twisted.web.client import HTTPClientFactory
from twisted.web import server, resource

#------------------------------------------------------------------------------

if __name__ == '__main__':
    import os.path as _p
    sys.path.insert(0, _p.abspath(_p.join(_p.dirname(_p.abspath(sys.argv[0])), '..')))
    sys.path.insert(0, _p.abspath(_p.join(_p.dirname(_p.abspath(sys.argv[0])), '..', '..')))

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.lib import nameurl
from bitdust.lib import net_misc
from bitdust.lib import strng

from bitdust.system import bpio
from bitdust.system import tmpfile

from bitdust.contacts import contactsdb
from bitdust.contacts import identitycache

from bitdust.userid import my_id

from bitdust.main import settings

#------------------------------------------------------------------------------

_Outbox = {}
_Contacts = {}
_Receiver = None
_ReceivingLoop = None
_ServerListener = None
_LastPingTimeDict = {}
_PingDelayDict = {}
_ConnectionsDict = {}
_CurrentDelay = 5

#------------------------------------------------------------------------------

# def init(receiving=True, sending_port=None):
#     """
#     """
#     lg.out(4, 'http_node.init')
#     contactsdb.AddContactsChangedCallback(on_contacts_changed)
#     if sending_port:
#         start_sending(sending_port)
#     if receiving:
#         start_receiving()

# def shutdown():
#     """
#     """
#     contactsdb.RemoveContactsChangedCallback(on_contacts_changed)
#     stop_sending()
#     stop_receiving()

#------------------------------------------------------------------------------


def start_sending(port):
    global _ServerListener

    if _ServerListener is not None:
        lg.out(8, 'http_node.start_http_server is already started')
        return _ServerListener

    lg.out(6, 'http_node.start_http_server going to listen on port ' + str(port))

    site = server.Site(SenderServer())
    try:
        _ServerListener = reactor.listenTCP(int(port), site)
    except:
        lg.exc()
        _ServerListener = None
    return _ServerListener


def stop_sending():
    global _ServerListener
    lg.out(6, 'http_node.stop_sending')
    if _ServerListener is None:
        lg.out(8, 'http_node.stop_sending _ServerListener is None')
        d = Deferred()
        d.callback('')
        return d
    d = _ServerListener.stopListening()
    _ServerListener = None
    return d


#------------------------------------------------------------------------------


def send_file(idurl, filename):
    lg.out(12, 'http_node.send to %s %s' % (idurl, filename))
    global _Outbox
    if idurl not in _Outbox:
        _Outbox[idurl] = []
    _Outbox[idurl].append(filename)
    #we want to keep only 10 last files.
    if len(_Outbox[idurl]) > 10:
        lostedfilename = _Outbox[idurl].pop(0)
        lg.warn('losted: "%s"' % lostedfilename)


#         transport_control.sendStatusReport(
#             'unknown',
#             lostedfilename,
#             'failed',
#             'http',)

#------------------------------------------------------------------------------


class SenderServer(resource.Resource):
    isLeaf = True

    def render_POST(self, request):
        global _Outbox
        idurl = request.getHeader('idurl')
        if idurl is None:
            return ''
        lg.out(14, 'http_node.SenderServer.render connection from ' + idurl)
        if idurl not in list(_Outbox.keys()):
            return ''
        r = ''
        for filename in _Outbox[idurl]:
            if not os.path.isfile(filename):
                continue
            if not os.access(filename, os.R_OK):
                continue
            src = bpio.ReadBinaryFile(filename)
            if src == '':
                continue
            src64 = base64.b64encode(src)
            r += src64 + '\n'
            lg.out(12, 'http_node.SenderServer.render sent %s to %s' % (filename, idurl))
            #TODO request.getPeer()
#             transport_control.sendStatusReport(
#                 request.getClient(),
#                 filename,
#                 'finished',
#                 'http',)
        _Outbox.pop(idurl, None)
        return r


#------------------------------------------------------------------------------


class TransportHTTPClientFactory(HTTPClientFactory):
    pass


class TransportHTTPProxyClientFactory(HTTPClientFactory):
    def setURL(self, url):
        HTTPClientFactory.setURL(self, url)
        self.path = url


#------------------------------------------------------------------------------


class Receiver(object):
    def loop(self):
        global _ReceivingLoop
        global _Contacts
        global _ToIncreaseDelay
        global _LastPingTimeDict
        global _PingDelayDict
        global _ConnectionsDict
        global _CurrentDelay
        lg.out(6, 'http_node.Receiver.loop')
        # _CurrentDelay = settings.getHTTPDelay()
        for idurl, hostport in _Contacts.items():
            if idurl in _ConnectionsDict:
                continue
            lasttm = _LastPingTimeDict.get(idurl, 0)
            delay = _PingDelayDict.get(idurl, _CurrentDelay)
            dt = time.time() - lasttm
            if dt < delay:
                continue
            _ConnectionsDict[idurl] = self.do_ping(idurl, hostport[0], hostport[1])
            _LastPingTimeDict[idurl] = time.time()
        _ReceivingLoop = reactor.callLater(1, self.loop)
        return _ReceivingLoop

    def on_ping_success(self, src, idurl, host, port, conn):
        global _LastPingTimeDict
        global _ConnectionsDict
        if len(src) == 0:
            increase_receiving_delay(idurl)
        else:
            parts = src.splitlines()
            lg.out(14, 'http_node.receive.success %d bytes in %d parts from %s (%s:%s)' % (len(src), len(parts), idurl, host, port))
            for part64 in parts:
                try:
                    part = base64.b64decode(part64.strip())
                except:
                    lg.out(14, 'http_node.receive.success ERROR in base64.b64decode()')
                    decrease_receiving_delay(idurl)
                    continue
                fd, filename = tmpfile.make('http-in', extension='.http')
                os.write(fd, part)
                os.close(fd)
                decrease_receiving_delay(idurl)
#                 transport_control.receiveStatusReport(
#                     filename,
#                     'finished',
#                     'http',
#                     host+':'+port,)
#             transport_control.log('http', 'finish connection with %s:%s ' % (host, port))
        conn.disconnect()
        # TODO: keep it opened!
        _ConnectionsDict.pop(idurl, None)

    def on_ping_failed(self, x, idurl, host, port, conn):
        global _LastPingTimeDict
        global _ConnectionsDict
        increase_receiving_delay(idurl)
        conn.disconnect()
        _ConnectionsDict.pop(idurl, None)

    def do_ping(self, idurl, host, port):
        lg.out(14, 'http_node.receive.ping     %s (%s:%s)' % (idurl, host, port))
        url = b'http://' + host + b':' + strng.to_bin(str(port))

        if net_misc.proxy_is_on():
            f = TransportHTTPProxyClientFactory(
                url,
                method='POST',
                headers={
                    'User-Agent': 'BitDust transport_http',
                    'idurl': my_id.getIDURL(),
                },
            )
            conn = reactor.connectTCP(net_misc.get_proxy_host(), int(net_misc.get_proxy_port()), f)
        else:
            f = TransportHTTPClientFactory(
                url,
                method='POST',
                headers={
                    'User-Agent': 'BitDust transport_http',
                    'idurl': my_id.getIDURL(),
                },
            )
            conn = reactor.connectTCP(host, int(port), f)

        f.deferred.addCallback(self.on_ping_success, idurl, host, port, conn)
        f.deferred.addErrback(self.on_ping_failed, idurl, host, port, conn)
        return conn


#------------------------------------------------------------------------------


def decrease_receiving_delay(idurl):
    global _PingDelayDict
    global _CurrentDelay
    lg.out(14, 'http_node.decrease_receiving_delay ' + idurl)
    _PingDelayDict[idurl] = _CurrentDelay


def increase_receiving_delay(idurl):
    global _PingDelayDict
    global _CurrentDelay
    if idurl not in _PingDelayDict:
        _PingDelayDict[idurl] = _CurrentDelay
    d = _PingDelayDict[idurl]
    if d < settings.DefaultSendTimeOutHTTP()/2:
        lg.out(14, 'http_node.increase_receiving_delay   %s for %s' % (str(d), idurl))
        _PingDelayDict[idurl] *= 2


#------------------------------------------------------------------------------


def start_receiving():
    global _Receiver
    if _Receiver is not None:
        lg.warn('already started')
        return _Receiver
    _Receiver = Receiver()
    _Receiver.loop()
    return _Receiver


def stop_receiving():
    global _ReceivingLoop
    lg.out(6, 'http_node.stop_receiving')
    if _ReceivingLoop is None:
        lg.out(8, 'http_node.stop_receiving _ReceivingLoop is None')
        return
    if _ReceivingLoop.called:
        lg.out(8, 'http_node.stop_receiving _ReceivingLoop is already called')
        return
    _ReceivingLoop.cancel()
    del _ReceivingLoop
    _ReceivingLoop = None


#------------------------------------------------------------------------------


def push_contact(idurl):
    global _Contacts
    global _PingDelayDict
    global _CurrentDelay
    ident = identitycache.FromCache(idurl)
    if ident is None:
        lg.err('"%s" not in the cache' % idurl)
        return None
    http_contact = ident.getProtoContact('http')
    if http_contact is None:
        if _Debug:
            lg.out(_DebugLevel*2, 'http_node.add_contact SKIP "%s" : no http contacts found in identity' % idurl)
        return None
    _, host, port, _ = nameurl.UrlParse(http_contact)
    new_item = False
    if idurl in _Contacts:
        new_item = True
    _Contacts[idurl] = (host, port)
    _PingDelayDict[idurl] = _CurrentDelay
    if new_item:
        if _Debug:
            lg.out(_DebugLevel, 'http_node.add_contact ADDED "%s" on %s:%s' % (idurl, host, port))
    else:
        if _Debug:
            lg.out(_DebugLevel, 'http_node.add_contact UPDATED "%s" on %s:%s' % (idurl, host, port))
    return idurl


#------------------------------------------------------------------------------


def do_update_contacts():
    global _Contacts
    if _Debug:
        lg.out(_DebugLevel, 'http_node.update_contacts')
    _Contacts.clear()

    for idurl in contactsdb.contacts(include_all=True):
        lg.out(10, 'http_node.update_contacts want ' + idurl)
        if idurl == my_id.getIDURL():
            continue
        latest_identity = identitycache.GetLatest(idurl)
        if isinstance(latest_identity, Deferred):
            latest_identity.addCallback(lambda src: push_contact(idurl))
            latest_identity.addErrback(lambda err: lg.out(_DebugLevel, 'http_node.update_contacts "%s" failed to cache' % idurl) if _Debug else None)
        else:
            push_contact(idurl)


#------------------------------------------------------------------------------


def on_contacts_changed(oldlist, newlist):
    do_update_contacts()


#------------------------------------------------------------------------------


def usage():
    print('''usage:
http_node.py send [server_port] [to idurl] [filename]
http_node.py receive
''')


def main():
    import logging
    logging.basicConfig(level=logging.DEBUG)
    from twisted.internet.defer import setDebugging
    setDebugging(True)
    # from twisted.python import log as twisted_log
    # twisted_log.startLogging(sys.stdout)
    lg.set_debug_level(20)
    settings.init()
    settings.update_proxy_settings()

    if sys.argv.count('receive'):
        start_receiving()
#         global _Contacts
#         _Contacts['http://p2p-id.ru/veselin.xml'] = ('127.0.0.1', 9122)

    elif sys.argv.count('send'):
        start_sending(port=int(sys.argv[2]))
        send_file(sys.argv[3], sys.argv[4])

    else:
        usage()
        return

    reactor.run()  # @UndefinedVariable
    settings.shutdown()


#------------------------------------------------------------------------------

if __name__ == '__main__':
    main()



#!/usr/bin/python
# p2p_queue.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (p2p_queue.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
.. module:: p2p_queue.


Methods to establish a messages queue between two or more nodes.:

    + Producers will send a messages to the queue
    + Consumers will listen to the queue and read the messages coming in
    + Producer only start sending if he have a Public Key
    + Consumer can only listen if he possess the correct Private Key
    + Queue is only stored on given node: both producer and consumer must be connected to that machine
    + Global queue ID is unique : queue_alias&alice@somehost.net&bob@anotherhost.com
    + Queue size is limited by a parameter, you can not publish when queue is overloaded

"""

#------------------------------------------------------------------------------

from __future__ import absolute_import
from __future__ import print_function

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 10

#------------------------------------------------------------------------------

import sys
import time

from collections import OrderedDict

#------------------------------------------------------------------------------

try:
    from twisted.internet import reactor  # @UnresolvedImport
except:
    sys.exit('Error initializing twisted.internet.reactor in p2p_queue.py')

from twisted.internet.defer import Deferred

#------------------------------------------------------------------------------

if __name__ == '__main__':
    import os.path as _p
    sys.path.insert(0, _p.abspath(_p.join(_p.dirname(_p.abspath(sys.argv[0])), '..')))

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.lib import utime
from bitdust.lib import misc
from bitdust.lib import packetid
from bitdust.lib import strng
from bitdust.lib import jsn
from bitdust.lib import serialization

from bitdust.main import events

from bitdust.crypt import my_keys
from bitdust.crypt import signed

from bitdust.p2p import commands
from bitdust.p2p import p2p_service

from bitdust.userid import global_id
from bitdust.userid import my_id
from bitdust.userid import id_url

#------------------------------------------------------------------------------

MAX_QUEUE_LENGTH = 100
MAX_CONSUMER_PENDING_MESSAGES = int(MAX_QUEUE_LENGTH/2)

MIN_PROCESS_QUEUES_DELAY = 0.1
MAX_PROCESS_QUEUES_DELAY = 2.0

#------------------------------------------------------------------------------

_ProcessQueuesDelay = 0.1
_ProcessQueuesTask = None
_ProcessQueuesLastTime = 0

_ActiveQueues = {}

_LastMessageID = None

_Producers = {}
_Consumers = {}

_EventPacketReceivedCallbacks = []
_MessageProcessedCallbacks = []

#------------------------------------------------------------------------------


def init():
    if _Debug:
        lg.out(_DebugLevel, 'p2p_queue.init')
    add_event_handler(do_handle_event_packet)
    start()


def shutdown():
    if _Debug:
        lg.out(_DebugLevel, 'p2p_queue.shutdown')
    remove_event_handler(do_handle_event_packet)
    stop()


#------------------------------------------------------------------------------


def make_message_id():
    """
    Generate a unique message ID to be stored in the queue.
    """
    global _LastMessageID
    if _LastMessageID is None:
        _LastMessageID = int(str(int(time.time()*100.0))[4:])
    _LastMessageID += 1
    return _LastMessageID


#------------------------------------------------------------------------------


def queue(queue_id=None):
    global _ActiveQueues
    if queue_id is None:
        return _ActiveQueues
    if queue_id not in _ActiveQueues:
        raise Exception('queue not found')
    return _ActiveQueues[queue_id]


def consumer(consumer_id=None):
    global _Consumers
    if consumer_id is None:
        return _Consumers
    if consumer_id not in _Consumers:
        raise Exception('consumer not found')
    return _Consumers[consumer_id]


def producer(producer_id=None):
    global _Producers
    if producer_id is None:
        return _Producers
    if producer_id not in _Producers:
        raise Exception('producer not found')
    return _Producers[producer_id]


#------------------------------------------------------------------------------


def start():
    if _Debug:
        lg.out(_DebugLevel, 'p2p_queue.start')
    reactor.callLater(0, process_queues)  # @UndefinedVariable
    return True


def stop():
    if _Debug:
        lg.out(_DebugLevel, 'p2p_queue.stop')
    global _ProcessQueuesTask
    if _ProcessQueuesTask:
        if _ProcessQueuesTask.active():
            _ProcessQueuesTask.cancel()
        _ProcessQueuesTask = None
        return True
    return False


def process_queues(interested_consumers=None):
    global _ProcessQueuesDelay
    global _ProcessQueuesTask
    global _ProcessQueuesLastTime
    has_activity = do_consume(interested_consumers=interested_consumers)
    _ProcessQueuesLastTime = time.time()
    if _ProcessQueuesTask is None or _ProcessQueuesTask.called:
        _ProcessQueuesDelay = misc.LoopAttenuation(
            _ProcessQueuesDelay,
            has_activity,
            MIN_PROCESS_QUEUES_DELAY,
            MAX_PROCESS_QUEUES_DELAY,
        )
        # attenuation
        _ProcessQueuesTask = reactor.callLater(_ProcessQueuesDelay, process_queues)  # @UndefinedVariable


def touch_queues(interested_consumers=None):
    global _ProcessQueuesDelay
    global _ProcessQueuesTask
    global _ProcessQueuesLastTime
    if time.time() - _ProcessQueuesLastTime < MIN_PROCESS_QUEUES_DELAY:
        return False
    reactor.callLater(0, process_queues, interested_consumers=interested_consumers)  # @UndefinedVariable
    return True


#------------------------------------------------------------------------------


def valid_queue_id(queue_id):
    if not queue_id:
        return False
    try:
        str(queue_id)
    except:
        return False
    queue_info = global_id.ParseGlobalQueueID(queue_id)
    if not misc.ValidName(queue_info['queue_alias']):
        return False
    owner_id = global_id.ParseGlobalID(queue_info['owner_id'])
    if not owner_id['idurl']:
        return False
    supplier_id = global_id.ParseGlobalID(queue_info['supplier_id'])
    if not supplier_id['idurl']:
        return False
    return True


def is_queue_exist(queue_id):
    return queue_id in queue()


def open_queue(queue_id):
    global _ActiveQueues
    if not valid_queue_id(queue_id):
        raise Exception('invalid queue id')
    if queue_id in queue():
        raise Exception('queue already exist')
    if _Debug:
        lg.args(_DebugLevel, queue_id=queue_id)
    _ActiveQueues[queue_id] = OrderedDict()
    lg.info('new queue opened: %s' % queue_id)
    return True


def close_queue(queue_id, remove_empty_consumers=False, remove_empty_producers=False):
    global _ActiveQueues
    if not valid_queue_id(queue_id):
        raise Exception('invalid queue id')
    if queue_id not in queue():
        raise Exception('queue not exist')
    if _Debug:
        lg.args(_DebugLevel, queue_id=queue_id)
    for producer_id in list(producer().keys()):
        if is_producer_connected(producer_id, queue_id):
            disconnect_producer(producer_id, queue_id, remove_empty=remove_empty_producers)
    for message_id in list(queue(queue_id).keys()):
        if message_id not in queue(queue_id):
            continue
        for consumer_id in list(queue(queue_id)[message_id].notifications.keys()):
            msg_obj = queue(queue_id).get(message_id)
            if msg_obj:
                callback_object = queue(queue_id)[message_id].notifications.get(consumer_id)
                if callback_object and not callback_object.called:
                    lg.info('canceling non-finished notification in the queue %s' % queue_id)
                    callback_object.cancel()
    for consumer_id in list(consumer().keys()):
        if is_consumer_subscribed(consumer_id, queue_id):
            unsubscribe_consumer(consumer_id, queue_id, remove_empty=remove_empty_consumers)
    _ActiveQueues.pop(queue_id)
    lg.info('existing queue closed: %s' % queue_id)
    return True


def rename_queue(old_queue_id, new_queue_id):
    if old_queue_id == new_queue_id:
        return False
    if not valid_queue_id(old_queue_id):
        raise Exception('invalid queue id')
    if not valid_queue_id(new_queue_id):
        raise Exception('invalid queue id')
    if old_queue_id not in queue():
        raise Exception('queue not exist')
    if _Debug:
        lg.args(_DebugLevel, old=old_queue_id, new=new_queue_id)
    stored_messages = queue().pop(old_queue_id)
    for message_id, msg_obj in stored_messages.items():
        for consumer_id, callback_object in list(msg_obj.notifications.items()):
            if not callback_object.called:
                callback_object.addCallback(on_notification_succeed, consumer_id, new_queue_id, message_id)
                callback_object.addErrback(on_notification_failed, consumer_id, new_queue_id, message_id)
    subscribed_consumers = list_subscribed_consumers(old_queue_id)
    connected_producers = list_connected_producers(old_queue_id)
    queue()[new_queue_id] = stored_messages
    for consumer_id in subscribed_consumers:
        consumer(consumer_id).queues.remove(old_queue_id)
        consumer(consumer_id).queues.append(new_queue_id)
    for producer_id in connected_producers:
        producer(producer_id).queues.remove(old_queue_id)
        producer(producer_id).queues.append(new_queue_id)
    lg.info('existing queue renamed: %s -> %s' % (old_queue_id, new_queue_id))
    return True


#------------------------------------------------------------------------------


def is_consumer_exists(consumer_id):
    return consumer_id in consumer()


def add_consumer(consumer_id):
    global _Consumers
    if consumer_id in consumer():
        raise Exception('consumer already exist')
    _Consumers[consumer_id] = ConsumerInfo(consumer_id)
    lg.info('new consumer added: %s' % consumer_id)
    return True


def remove_consumer(consumer_id):
    global _Consumers
    if consumer_id not in consumer():
        raise Exception('consumer not exist')
    old_consumer = _Consumers.pop(consumer_id)
    old_consumer.commands = {}
    old_consumer.queues = []
    lg.info('existing consumer removed: %s' % str(consumer_id))
    return True


#------------------------------------------------------------------------------


def is_callback_method_registered(consumer_id, callback_method):
    if consumer_id not in consumer():
        return False
    if callback_method not in consumer(consumer_id).commands:
        return False
    return True


def add_callback_method(consumer_id, callback_method, interested_queues_list=None):
    if consumer_id not in consumer():
        raise Exception('consumer not found')
    if callback_method in consumer(consumer_id).commands:
        raise Exception('callback method already exist')
    consumer(consumer_id).commands[callback_method] = interested_queues_list
    if _Debug:
        lg.args(_DebugLevel, c=consumer_id, cb=callback_method)
    return True


def remove_callback_method(consumer_id, callback_method):
    if consumer_id not in consumer():
        raise Exception('consumer not found')
    if callback_method not in consumer(consumer_id).commands:
        raise Exception('callback method not found')
    consumer(consumer_id).commands.pop(callback_method)
    if _Debug:
        lg.args(_DebugLevel, c=consumer_id, cb=callback_method)
    return True


#------------------------------------------------------------------------------


def is_consumer_subscribed(consumer_id, queue_id=None):
    if queue_id and not valid_queue_id(queue_id):
        return False
    if consumer_id not in consumer():
        return False
    if not queue_id:
        return len(consumer(consumer_id).queues) > 0
    if queue_id not in consumer(consumer_id).queues:
        return False
    return True


def list_subscribed_consumers(queue_id):
    if not valid_queue_id(queue_id):
        return []
    return [c_id for c_id in consumer().keys() if queue_id in consumer(c_id).queues]


def subscribe_consumer(consumer_id, queue_id):
    if not valid_queue_id(queue_id):
        raise Exception('invalid queue id')
    if consumer_id not in consumer():
        raise Exception('consumer not found')
    if queue_id in consumer(consumer_id).queues:
        raise Exception('consumer is already subscribed')
    consumer(consumer_id).queues.append(queue_id)
    lg.info('consumer %s subscribed to read queue %s' % (consumer_id, queue_id))
    return True


def unsubscribe_consumer(consumer_id, queue_id=None, remove_empty=False):
    if not valid_queue_id(queue_id):
        raise Exception('invalid queue id')
    if consumer_id not in consumer():
        raise Exception('consumer not found')
    if queue_id is None:
        consumer(consumer_id).queues = []
        lg.info('consumer %s unsubscribed from all queues' % (consumer_id, ))
        if remove_empty:
            remove_consumer(consumer_id)
        return True
    if queue_id not in consumer(consumer_id).queues:
        raise Exception('consumer is not subscribed')
    consumer(consumer_id).queues.remove(queue_id)
    lg.info('consumer %s unsubscribed from queue %s' % (consumer_id, queue_id))
    if remove_empty:
        if len(consumer(consumer_id).queues) == 0:
            remove_consumer(consumer_id)
    return True


#------------------------------------------------------------------------------


def is_producer_exist(producer_id):
    return producer_id in producer()


def add_producer(producer_id):
    global _Producers
    if is_producer_exist(producer_id):
        raise Exception('producer already exist')
    _Producers[producer_id] = ProducerInfo(producer_id)
    lg.info('new producer added: %s' % producer_id)
    return True


def remove_producer(producer_id):
    global _Producers
    if not is_producer_exist(producer_id):
        raise Exception('producer not exist')
    old_producer = _Producers.pop(producer_id)
    old_producer.queues = []
    for event_id in list(old_producer.publishers.keys()):
        old_producer.stop_publisher(event_id)
    lg.info('existing producer removed: %s' % str(producer_id))
    return True


#------------------------------------------------------------------------------


def is_producer_connected(producer_id, queue_id=None):
    if queue_id and not valid_queue_id(queue_id):
        return False
    if not is_producer_exist(producer_id):
        return False
    if not queue_id:
        return len(producer(producer_id).queues) > 0
    return queue_id in producer(producer_id).queues


def list_connected_producers(queue_id):
    if not valid_queue_id(queue_id):
        return []
    return [p_id for p_id in producer().keys() if queue_id in producer(p_id).queues]


def connect_producer(producer_id, queue_id):
    if not is_producer_exist(producer_id):
        raise Exception('producer not exist')
    if not is_queue_exist(queue_id):
        raise Exception('queue not exist')
    if queue_id not in producer(producer_id).queues:
        producer(producer_id).queues.append(queue_id)
        lg.info('producer %s connected to queue %s' % (producer_id, queue_id))
    else:
        lg.warn('producer %s already connected to queue %s' % (producer_id, queue_id))
    return True


def disconnect_producer(producer_id, queue_id=None, remove_empty=False):
    if not is_producer_exist(producer_id):
        raise Exception('producer not exist')
    if not is_queue_exist(queue_id):
        raise Exception('queue not exist')
    if queue_id is None:
        producer(producer_id).queues = []
        lg.info('producer %s disconnected from all queues' % (producer_id, ))
        if remove_empty:
            remove_producer(producer_id)
        return True
    if queue_id not in producer(producer_id).queues:
        raise Exception('producer is not connected to that queue')
    producer(producer_id).queues.remove(queue_id)
    lg.info('producer %s disconnected from queue %s' % (producer_id, queue_id))
    if remove_empty:
        if len(producer(producer_id).queues) == 0:
            remove_producer(producer_id)
    return True


#------------------------------------------------------------------------------


def is_event_publishing(producer_id, event_id):
    if not is_producer_exist(producer_id):
        return False
    return producer(producer_id).is_event_publishing(event_id)


def start_event_publisher(producer_id, event_id):
    if not is_producer_exist(producer_id):
        raise Exception('producer not exist')
    return producer(producer_id).start_publisher(event_id)


def stop_event_publisher(producer_id, event_id):
    if not is_producer_exist(producer_id):
        raise Exception('producer not exist')
    return producer(producer_id).stop_publisher(event_id)


#------------------------------------------------------------------------------


def start_notification(consumer_id, queue_id, message_id):
    if not valid_queue_id(queue_id):
        raise Exception('invalid queue id')
    if consumer_id not in consumer():
        raise Exception('consumer not found')
    if queue_id not in queue():
        raise Exception('queue not exist')
    if message_id not in queue(queue_id):
        raise Exception('message not exist')
    if consumer_id in queue(queue_id)[message_id].notifications:
        raise Exception('notification already sent to given consumer')
    callback_object = Deferred()
    queue(queue_id)[message_id].notifications[consumer_id] = callback_object
    consumer(consumer_id).consumed_messages += 1
    callback_object.addCallback(on_notification_succeed, consumer_id, queue_id, message_id)
    callback_object.addErrback(on_notification_failed, consumer_id, queue_id, message_id)
    queue(queue_id)[message_id].state = 'SENT'
    if _Debug:
        lg.args(_DebugLevel, consumer_id=consumer_id, queue_id=queue_id, message_id=message_id, notifications=len(queue(queue_id)[message_id].notifications))
    return callback_object


def finish_notification(consumer_id, queue_id, message_id, success):
    if not valid_queue_id(queue_id):
        raise Exception('invalid queue id')
    if queue_id not in queue():
        raise Exception('queue not exist')
    if message_id not in queue(queue_id):
        raise Exception('message not exist')
    if consumer_id not in queue(queue_id)[message_id].notifications:
        raise Exception('not found pending notification for given consumer')
    defer_result = queue(queue_id)[message_id].notifications[consumer_id]
    if defer_result is None:
        raise Exception('notification already finished')
    if not isinstance(defer_result, Deferred):
        raise Exception('invalid notification type')
    queue(queue_id)[message_id].notifications[consumer_id] = None
    # queue(queue_id)[message_id].notifications.pop(consumer_id)
    if success:
        queue(queue_id)[message_id].success_notifications.append(consumer_id)
        consumer(consumer_id).success_notifications += 1
    else:
        queue(queue_id)[message_id].failed_notifications.append(consumer_id)
        consumer(consumer_id).failed_notifications += 1
    if not defer_result.called:
        lg.info('canceling non-finished notification in the queue %s' % queue_id)
        defer_result.cancel()
    del defer_result
    if _Debug:
        lg.args(_DebugLevel, consumer_id=consumer_id, queue_id=queue_id, message_id=message_id, success=success, notifications=len(queue(queue_id)[message_id].notifications))
    return True


#------------------------------------------------------------------------------


def on_notification_succeed(result, consumer_id, queue_id, message_id):
    if _Debug:
        lg.out(_DebugLevel, 'p2p_queue.on_notification_succeed : message %r delivered to consumer %r in %r' % (message_id, consumer_id, queue_id))
    if is_queue_exist(queue_id) and is_consumer_exists(consumer_id):
        try:
            # reactor.callLater(0, finish_notification, consumer_id, queue_id, message_id, success=True)  # @UndefinedVariable
            finish_notification(consumer_id, queue_id, message_id, success=True)
        except:
            lg.exc()
    else:
        lg.warn('notification %r was not finished for consumer %r in %r' % (message_id, consumer_id, queue_id))
    # TODO: add a counter and execute cleanup less frequently
    do_cleanup(target_queues=[queue_id])
    # reactor.callLater(0, do_cleanup, target_queues=[queue_id, ])  # @UndefinedVariable
    return result


def on_notification_failed(err, consumer_id, queue_id, message_id):
    if _Debug:
        lg.out(_DebugLevel, 'p2p_queue.on_notification_failed FAILED message %r for consumer %r in %r : %r' % (message_id, consumer_id, queue_id, err.getErrorMessage()))
    if is_queue_exist(queue_id) and is_consumer_exists(consumer_id):
        try:
            # reactor.callLater(0, finish_notification, consumer_id, queue_id, message_id, success=False)  # @UndefinedVariable
            finish_notification(consumer_id, queue_id, message_id, success=False)
        except:
            lg.exc()
    else:
        lg.warn('failed notification %r was not finished for consumer %r in %r' % (message_id, consumer_id, queue_id))
    # TODO: add a counter and execute cleanup less frequently
    do_cleanup(target_queues=[queue_id])
    return None


#------------------------------------------------------------------------------


def write_message(producer_id, queue_id, data, creation_time=None):
    if not valid_queue_id(queue_id):
        raise Exception('invalid queue id')
    if not is_producer_exist(producer_id):
        raise Exception('unknown producer')
    if not is_producer_connected(producer_id, queue_id):
        raise Exception('producer was not connected to the queue')
    if len(queue(queue_id)) >= MAX_QUEUE_LENGTH:
        raise P2PQueueIsOverloaded('queue is overloaded')
    producer(producer_id).produced_messages += 1
    new_message = QueueMessage(producer_id, queue_id, data, created=creation_time)
    queue(queue_id)[new_message.message_id] = new_message
    queue(queue_id)[new_message.message_id].state = 'PUSHED'
    if _Debug:
        lg.out(_DebugLevel, 'p2p_queue.write_message  %r added to queue %s' % (new_message.message_id, queue_id))
    touch_queues()
    return new_message


def pull_message(queue_id, message_id=None):
    if not valid_queue_id(queue_id):
        raise Exception('invalid queue id')
    if queue_id not in list(queue().keys()):
        raise Exception('queue id not found')
    if message_id is None:
        if len(list(queue(queue_id).keys())) == 0:
            lg.info('there is no messages in the queue %s' % queue_id)
            return None
        message_id = list(queue(queue_id).keys())[0]
    if message_id not in queue(queue_id):
        lg.info('given message was not found in the queue %s' % queue_id)
        return None
    existing_message = queue(queue_id).pop(message_id)
    existing_message.state = 'PULLED'
    if _Debug:
        lg.out(_DebugLevel, 'p2p_queue.pull_message  %r removed from queue %s' % (message_id, queue_id))
    return existing_message


def lookup_pending_message(consumer_id, queue_id):
    if not valid_queue_id(queue_id):
        raise Exception('invalid queue id')
    if queue_id not in queue():
        raise Exception('queue not exist')
    if consumer_id not in consumer():
        raise Exception('consumer not found')
    pending_message_id = None
    # here we assume that OrderedDict is really ordered
    for message_id, message_obj in queue(queue_id).items():
        # loop all messages from the beginning
        if consumer_id not in message_obj.consumers:
            # only interested consumers needs to be selected
            continue
        if consumer_id in message_obj.success_notifications or consumer_id in message_obj.failed_notifications:
            # only select consumer which was not notified yet
            continue
        if consumer_id in message_obj.notifications:
            # notification already started to given consumer
            continue
        pending_message_id = message_id
        break
    return pending_message_id


#------------------------------------------------------------------------------


def push_signed_message(producer_id, queue_id, data, creation_time=None):
    # TODO: to be continue
    try:
        signed_data = signed.Packet(
            Command=commands.Event(),
            OwnerID=producer_id,
            CreatorID=my_id.getIDURL(),
            PacketID=packetid.UniqueID(),
            Payload=serialization.DictToBytes(data, keys_to_text=True),
            RemoteID=queue_id,
            KeyID=producer_id,
        )
    except:
        lg.exc()
        raise Exception('sign message failed')
    return write_message(producer_id, queue_id, data=signed_data.Serialize(), creation_time=creation_time)


def pop_signed_message(queue_id, message_id):
    # TODO: to be continue
    existing_message = pull_message(queue_id, message_id)
    if not existing_message:
        return existing_message
    try:
        signed_data = signed.Unserialize(existing_message.payload)
    except:
        raise Exception('unserialize message fails')
    if not signed_data:
        raise Exception('unserialized message is empty')
    if not signed_data.Valid():
        raise Exception('unserialized message is not valid')
    try:
        existing_message.payload = serialization.BytesToDict(signed_data.Payload, keys_to_text=True)
    except:
        raise Exception('failed reading message json data')
    return existing_message


#------------------------------------------------------------------------------


def add_event_handler(cb):
    global _EventPacketReceivedCallbacks
    _EventPacketReceivedCallbacks.append(cb)


def insert_event_handler(cb):
    global _EventPacketReceivedCallbacks
    _EventPacketReceivedCallbacks.insert(0, cb)


def remove_event_handler(cb):
    global _EventPacketReceivedCallbacks
    _EventPacketReceivedCallbacks.remove(cb)


#------------------------------------------------------------------------------


def add_message_processed_callback(cb):
    global _MessageProcessedCallbacks
    _MessageProcessedCallbacks.append(cb)


def remove_message_processed_callback(cb):
    global _MessageProcessedCallbacks
    _MessageProcessedCallbacks.remove(cb)


#------------------------------------------------------------------------------


def on_event_packet_received(newpacket, info, status, error_message):
    global _EventPacketReceivedCallbacks
    try:
        e_json = serialization.BytesToDict(newpacket.Payload, keys_to_text=True, values_to_text=True)
        strng.to_text(e_json['event_id'])
    except:
        lg.warn('invalid json payload')
        return False
    handled = False
    for cb in _EventPacketReceivedCallbacks:
        handled = cb(newpacket, e_json)
        if handled:
            break
    return handled


def do_handle_event_packet(newpacket, e_json):
    event_id = strng.to_text(e_json['event_id'])
    payload = e_json['payload']
    queue_id = strng.to_text(e_json.get('queue_id'))
    producer_id = e_json.get('producer_id')
    message_id = strng.to_text(e_json.get('message_id'))
    created = strng.to_text(e_json.get('created'))
    if _Debug:
        lg.args(_DebugLevel, event_id=event_id, queue_id=queue_id, producer_id=producer_id, message_id=message_id)
    if queue_id and producer_id and message_id:
        # this message have an ID and producer so it came from a queue and needs to be consumed
        # also, more info coming from the queue needs to be attached to the event body
        # TODO: need more verifications to be implemented here
        # SECURITY
        if _Debug:
            lg.info('received new event %s from the queue at %s' % (event_id, queue_id))
        payload.update(dict(
            queue_id=queue_id,
            producer_id=producer_id,
            message_id=message_id,
            created=created,
        ))
        events.send(event_id, data=payload)
        p2p_service.SendAck(newpacket)
        return True
    if producer_id == my_id.getID() and not queue_id:
        # this message addressed directly to me, not to any specific queue
        return True
    # this message does not have nor ID nor producer so it came from another user directly
    # lets' try to find a queue for that event and see if we need to publish it or not
    queue_id = global_id.MakeGlobalQueueID(
        queue_alias=event_id,
        owner_id=global_id.MakeGlobalID(idurl=newpacket.OwnerID),
        supplier_id=global_id.MakeGlobalID(idurl=my_id.getGlobalID()),
    )
    if queue_id not in queue():
        # such queue is not found locally, that means message is
        # probably addressed to that node and needs to be consumed directly
        # TODO: check if we actually should consume the event locally and populate the event
        # need more verifications to be implemented here
        # SECURITY
        if _Debug:
            lg.warn('received event %s was not delivered to any queue, consume now and send an Ack' % event_id)
        # also add more info coming from the queue
        payload.update(dict(
            queue_id=queue_id,
            producer_id=producer_id,
            message_id=message_id,
            created=created,
        ))
        events.send(event_id, data=payload)
        p2p_service.SendAck(newpacket)
        return True
    # found a queue for that message, pushing there
    # TODO: add verification of producer's identity and signature
    # SECURITY
    if _Debug:
        lg.info('pushing event %s to the queue %s on behalf of producer %s' % (event_id, queue_id, producer_id))
    try:
        write_message(
            producer_id=producer_id,
            queue_id=queue_id,
            data=payload,
            creation_time=created,
        )
    except Exception as exc:
        lg.exc()
        p2p_service.SendFail(newpacket, str(exc))
        return True
    p2p_service.SendAck(newpacket)
    return True


#------------------------------------------------------------------------------


def do_notify(callback_method, consumer_id, queue_id, message_id):
    existing_message = queue(queue_id)[message_id]
    event_id = global_id.ParseGlobalQueueID(queue_id)['queue_alias']
    if consumer_id in existing_message.notifications:
        if _Debug:
            lg.dbg(_DebugLevel, 'notification %r already started for consumer %r' % (message_id, consumer_id))
        # notification already sent to given consumer
        return False
    if _Debug:
        lg.args(_DebugLevel, cb=callback_method, c=consumer_id, q=queue_id, m=message_id)
    ret = start_notification(consumer_id, queue_id, message_id)
    if id_url.is_idurl(callback_method):
        p2p_service.SendEvent(
            remote_idurl=id_url.field(callback_method),
            event_id=event_id,
            payload=existing_message.payload,
            producer_id=existing_message.producer_id,
            consumer_id=consumer_id,
            queue_id=queue_id,
            message_id=existing_message.message_id,
            created=existing_message.created,
            callbacks={
                commands.Ack(): lambda response, info: ret.callback(True),
                commands.Fail(): lambda response, info: ret.callback(False),
                None: lambda pkt_out: ret.callback(False),
            },
        )
    else:
        try:
            result = callback_method(
                dict(
                    event_id=event_id,
                    payload=existing_message.payload,
                    producer_id=existing_message.producer_id,
                    consumer_id=consumer_id,
                    queue_id=queue_id,
                    message_id=existing_message.message_id,
                    created=existing_message.created,
                )
            )
        except:
            lg.exc('%r %r %r %r' % (callback_method, consumer_id, queue_id, message_id))
            result = False
        if isinstance(result, Deferred):
            result.addCallback(lambda ok: ret.callback(True) if ok else ret.callback(False))
            result.addErrback(lg.errback, debug=_Debug, debug_level=_DebugLevel, method='p2p_queue.do_notify')
            result.addErrback(lambda err: ret.callback(False))
        else:
            reactor.callLater(0, ret.callback, result)  # @UndefinedVariable
    return ret


def do_consume(interested_consumers=None):
    if not interested_consumers:
        interested_consumers = list(consumer().keys())
    disconnected_consumers = []
    to_be_consumed = []
    for consumer_id in interested_consumers:
        if len(consumer(consumer_id).queues) == 0:
            # skip, consumer is not subscribed to any queues
            continue
        if len(consumer(consumer_id).commands) == 0:
            # skip, no available notification methods found for given consumer
            disconnected_consumers.append(consumer_id)
            continue
        interested_queues = set()
        for queue_id in consumer(consumer_id).queues:
            if queue_id not in queue():
                continue
            if len(queue(queue_id)) == 0:
                # no messages in the queue
                continue
            interested_queues.add(queue_id)
        if len(interested_queues) == 0:
            # skip, no new messages in the queues which consumer subscribed on
            continue
        for queue_id in interested_queues:
            to_be_consumed.append((
                consumer_id,
                queue_id,
            ))
    if not to_be_consumed:
        # nothing to consume
        return False
    notifications_count = 0
    consumers_affected = []
    for _consumer_id, _queue_id in to_be_consumed:
        if _consumer_id in consumers_affected:
            # only one message per consumer at a time
            continue
        _message_id = lookup_pending_message(_consumer_id, _queue_id)
        if _message_id is None:
            # no new messages found for that consumer
            continue
        for callback_method, interested_queues_list in consumer(_consumer_id).commands.items():
            if interested_queues_list:
                matching = False
                for interested_queue in interested_queues_list:
                    if _queue_id.startswith(interested_queue):
                        matching = True
                        break
                if not matching:
                    continue
            do_notify(callback_method, _consumer_id, _queue_id, _message_id)
            notifications_count += 1
            consumers_affected.append(_consumer_id)
            break
    if _Debug:
        lg.args(_DebugLevel, notifications_count=notifications_count, consumers_affected=consumers_affected)
    del to_be_consumed
    del consumers_affected
    if notifications_count == 0:
        # nothing was sent
        return False
    return True


def do_cleanup(target_queues=None):
    global _MessageProcessedCallbacks
    to_be_removed = set()
    if not target_queues:
        target_queues = list(queue().keys())
    for queue_id in target_queues:
        if not is_queue_exist(queue_id):
            continue
        for _message in queue(queue_id).values():
            if _message.state == 'SENT':
                found_pending_notifications = False
                for defer_result in _message.notifications.values():
                    if defer_result and not defer_result.called:
                        found_pending_notifications = True
                if not found_pending_notifications:
                    # no pending notifications found, but state is SENT : all is done
                    to_be_removed.add((queue_id, _message.message_id))
                    continue
                if len(_message.failed_notifications) + len(_message.success_notifications) >= len(_message.consumers):
                    # all notifications are sent and results are received (or timeouts) - remove message from the queue
                    to_be_removed.add((queue_id, _message.message_id))
                    continue
            if len(_message.consumers) == 0:
                # there is no consumers for that message - remove it
                to_be_removed.add((queue_id, _message.message_id))
                continue
    for queue_id, message_id in to_be_removed:
        processed_message = pull_message(queue_id, message_id)
        if processed_message:
            for cb in _MessageProcessedCallbacks:
                if not cb(processed_message):
                    lg.warn('message %r was not correctly processed in %r' % (message_id, cb))
    to_be_removed.clear()
    del to_be_removed
    return True


#------------------------------------------------------------------------------


class QueueMessage(object):

    def __init__(self, producer_id, queue_id, json_data, created=None):
        self.message_id = make_message_id()
        self.producer_id = producer_id
        self.queue_id = queue_id
        self.created = created or utime.utcnow_to_sec1970()
        self.payload = jsn.dict_items_to_text(json_data)
        self.state = 'CREATED'
        self.notifications = {}
        self.success_notifications = []
        self.failed_notifications = []
        self.consumers = []
        for consumer_id in consumer():
            if queue_id in consumer(consumer_id).queues:
                self.consumers.append(consumer_id)
        if len(self.consumers) == 0:
            if _Debug:
                lg.warn('message %r from %r in queue %r will have no consumers' % (self.message_id, self.producer_id, self.queue_id))

    def __repr__(self):
        return 'QueueMessage[%s](%s by %s in %s)' % (
            self.state,
            self.message_id,
            self.producer_id,
            self.queue_id,
        )

    def get_sequence_id(self):
        return self.payload.get('sequence_id', None)


#------------------------------------------------------------------------------


class ConsumerInfo(object):

    def __init__(self, consumer_id):
        self.state = 'READY'
        self.consumer_id = consumer_id
        self.commands = {}
        self.queues = []
        self.consumed_messages = 0
        self.success_notifications = 0
        self.failed_notifications = 0


#------------------------------------------------------------------------------


class ProducerInfo(object):

    def __init__(self, producer_id):
        self.state = 'READY'
        self.producer_id = producer_id
        self.produced_messages = 0
        self.queues = []
        self.publishers = {}

    def is_event_publishing(self, event_id):
        return event_id in self.publishers

    def do_push_message(self, evt):
        if not self.queues:
            if _Debug:
                lg.warn('producer is not connected to any queue')
            return False
        for queue_id in self.queues:
            try:
                write_message(producer_id=self.producer_id, queue_id=queue_id, data=evt.data, creation_time=evt.created)
            except P2PQueueIsOverloaded as exc:
                lg.warn('queue_id=%s producer_id=%s: %s' % (queue_id, self.producer_id, exc))
        return True

    def start_publisher(self, event_id):
        if event_id in self.publishers:
            raise Exception('event publisher already exist')
        self.publishers[event_id] = lambda evt: self.do_push_message(evt)
        return events.add_subscriber(self.publishers[event_id], event_id)

    def stop_publisher(self, event_id):
        if event_id not in self.publishers:
            raise Exception('event publisher not found')
        result = events.remove_subscriber(self.publishers[event_id], event_id)
        if not self.publishers.pop(event_id, None):
            lg.warn('publisher event %s for producer %s was not cleaned correctly' % (event_id, self.producer_id))
        if not result:
            lg.warn('event subscriber for %s was not removed' % event_id)
        return result


#------------------------------------------------------------------------------


class P2PQueueIsOverloaded(Exception):
    pass


#------------------------------------------------------------------------------


def _test_callback_bob(message_json):
    # time.sleep(1)
    print('               !!!!!!!!!!!!! _test_callback_bob:', message_json)
    return True


def _test_callback_dave(message_json):
    # time.sleep(1)
    print('               !!!!!!!!!!!!! _test_callback_dave:', message_json)
    return True


def test():
    lg.set_debug_level(24)
    init()
    my_keys.generate_key('customer$bob@server-second.com')
    open_queue('event-test123&bob@server-second.com&carl@thirdnode.net')
    add_producer('alice@host-one.com')
    connect_producer('alice@host-one.com', 'event-test123&bob@server-second.com&carl@thirdnode.net')
    start_event_publisher('alice@host-one.com', 'test123')
    add_consumer('bob@server-second.com')
    add_callback_method('bob@server-second.com', _test_callback_bob)
    subscribe_consumer('bob@server-second.com', 'event-test123&bob@server-second.com&carl@thirdnode.net')
    add_consumer('dave@server-4.com')
    add_callback_method('dave@server-4.com', _test_callback_dave)
    subscribe_consumer('dave@server-4.com', 'event-test123&bob@server-second.com&carl@thirdnode.net')
    events.send('test123', data=dict(abc='abc', counter=0))
    events.send('test123', data=dict(abc='abc', counter=2))
    events.send('test123', data=dict(abc='abc', counter=4))
    events.send('test123', data=dict(abc='abc', counter=1))
    events.send('test123', data=dict(abc='abc', counter=6))
    events.send('test123', data=dict(abc='abc', counter=23))
    events.send('test123', data=dict(abc='abc', counter=620))
    events.send('test123', data=dict(abc='abc', counter=20))
    events.send('test123', data=dict(abc='abc', counter=30))
    events.send('test123', data=dict(abc='abc', counter=40))
    reactor.run()  # @UndefinedVariable


if __name__ == '__main__':
    test()
